"""URLParser

URL parsing is done through objects which are subclasses of the `URLParser`
class. `Application` delegates most of the URL parsing to these objects.

Application has a single "root" URL parser, which is used to parse all URLs.
This parser then can pass the request on to other parsers, usually taking off
parts of the URL with each step.

This root parser is generally `ContextParser`, which is instantiated and set
up by `Application` (accessible through `Application.rootURLParser`).
"""

import os
import re
import sys

from collections import defaultdict
from urllib.parse import unquote
from warnings import warn

from HTTPExceptions import HTTPNotFound, HTTPMovedPermanently
from MiscUtils.ParamFactory import ParamFactory

debug = False

# Legal characters for use in a module name -- used when turning
# an entire path into a module name.
_moduleNameRE = re.compile('[^a-zA-Z_]')

_globalApplication = None


def application():
    """Returns the global Application."""
    return _globalApplication


class URLParser:
    """URLParser is the base class for all URL parsers.

    Though its functionality is sparse, it may be expanded in the future.
    Subclasses should implement a `parse` method, and may also want to
    implement an `__init__` method with arguments that control how the
    parser works (for instance, passing a starting path for the parser)

    The `parse` method is where most of the work is done. It takes two
    arguments -- the transaction and the portion of the URL that is still to
    be parsed. The transaction may (and usually is) modified along the way.
    The URL is passed through so that you can take pieces off the front,
    and then pass the reduced URL to another parser. The method should return
    a servlet (never None).

    If you cannot find a servlet, or some other (somewhat) expected error
    occurs, you should raise an exception. HTTPNotFound probably being the
    most interesting.
    """

    def findServletForTransaction(self, trans):
        """Returns a servlet for the transaction.

        This is the top-level entry point, below it `parse` is used.
        """
        # pylint: disable=no-member
        return self.parse(trans, trans.request().urlPath())


class ContextParser(URLParser):
    """Find the context of a request.

    ContextParser uses the ``Application.config`` context settings to find
    the context of the request.  It then passes the request to a FileParser
    rooted in the context path.

    The context is the first element of the URL, or if no context matches
    that then it is the ``default`` context (and the entire URL is passed
    to the default context's FileParser).

    There is generally only one ContextParser, which can be found as
    ``application.rootURLParser()``.
    """

    # region Init

    def __init__(self, app):
        """Create ContextParser.

        ContextParser is usually created by Application, which
        passes all requests to it.

        In __init__ we take the ``Contexts`` setting from
        Application.config and parse it slightly.
        """
        URLParser.__init__(self)
        # Need to set this here because we need this during initialization:
        self._app = app
        self._imp = app._imp
        # self._context will be a dictionary of context names and context
        # directories.  It is set by `addContext`.
        self._contexts = {}
        # add all contexts except the default, which we save until the end
        contexts = app.setting('Contexts')
        defaultContext = ''
        for name, path in contexts.items():
            path = os.path.normpath(path)  # for Windows
            if name == 'default':
                defaultContext = path
            else:
                name = '/'.join(p for p in name.split('/') if p)
                self.addContext(name, path)
        if not defaultContext:
            # If no default context has been specified, and there is a unique
            # context not built into Webware, use it as the default context.
            for name in contexts:
                if name in ('Admin', 'Examples', 'Testing'):
                    continue
                if defaultContext:
                    defaultContext = None
                    break
                defaultContext = name
            if not defaultContext:
                # otherwise, try using the following contexts if available
                for defaultContext in ('Default', 'Examples'):
                    if defaultContext in contexts:
                        break
                else:  # if not available, refuse the temptation to guess
                    raise KeyError("No default context has been specified.")
        if defaultContext in self._contexts:
            self._defaultContext = defaultContext
        else:
            for name, path in self._contexts.items():
                if defaultContext == path:
                    self._defaultContext = name
                    break
            else:
                self.addContext('default', defaultContext)
                self._defaultContext = 'default'

    # endregion Init

    # region Context handling

    def resolveDefaultContext(self, dest):
        """Find default context.

        Figure out if the default context refers to an existing context,
        the same directory as an existing context, or a unique directory.

        Returns the name of the context that the default context refers to,
        or 'default' if the default context is unique.
        """
        contexts = self._contexts
        if dest in contexts:
            # The default context refers to another context,
            # not a unique context.  Return the name of that context.
            return dest
        destPath = self.absContextPath(dest)
        for name, path in contexts.items():
            if name != 'default' and self.absContextPath(path) == destPath:
                # The default context has the same directory
                # as another context, so it's still not unique
                return name
        # The default context has no other name
        return 'default'

    def addContext(self, name, path):
        """Add a context to the system.

        The context will be imported as a package, going by `name`,
        from the given directory path. The directory doesn't have to match
        the context name.
        """
        if name == 'default':
            dest = self.resolveDefaultContext(name)
            self._defaultContext = dest
            if dest != 'default':
                # in this case default refers to an existing context, so
                # there's not much to do
                print('Default context aliases to:', dest)
                return

        if name in sys.modules:
            module = sys.modules[name]
        else:
            try:
                localDir, packageName = os.path.split(path)
                spec = self._imp.findSpec(packageName, localDir or '.')
                try:
                    module = self._imp.moduleFromSpec(spec)
                except ImportError as e:
                    # Maybe this happened because it had been forgotten
                    # to add the __init__.py file. So we try to create one:
                    if os.path.exists(path):
                        initPy = os.path.join(path, '__init__.py')
                        if (not os.path.exists(initPy) and
                                not os.path.exists(initPy + 'c')):
                            print("Creating __init__.py file"
                                  f" for context {name!r}")
                            try:
                                with open(initPy, 'w') as initPyFile:
                                    initPyFile.write(
                                        '# Auto-generated by Webware' +
                                        os.linesep)
                            except Exception:
                                print("Error: __init__.py file"
                                      " could not be created.")
                            else:
                                module = self._imp.moduleFromSpec(spec)
                                e = None
                    if e:
                        raise
                else:
                    sys.modules[name] = module
            except ImportError as e:
                error = str(e) or 'Could not import package'
                print(f'Error loading context: {name}: {error}: dir={path}')
                return

        if hasattr(module, 'contextInitialize'):
            result = module.contextInitialize(
                application(),
                os.path.normpath(os.path.join(os.getcwd(), path)))
            if result is not None and 'ContentLocation' in result:
                path = result['ContentLocation']

        print(f'Loading context: {name} at {path}')
        self._contexts[name] = path

    def absContextPath(self, path):
        """Get absolute context path.

        Resolves relative paths, which are assumed to be relative to the
        Application's serverSidePath (the working directory).
        """
        if os.path.isabs(path):
            return path
        return self._app.serverSidePath(path)

    # endregion Context handling

    # region Parsing

    def parse(self, trans, requestPath):
        """Parse request.

        Get the context name, and dispatch to a FileParser rooted
        in the context's path.

        The context name and file path are stored in the request (accessible
        through `Request.serverSidePath` and `Request.contextName`).
        """
        # This is a hack... should probably go in the Transaction class:
        trans._fileParserInitSeen = defaultdict(set)
        # If there is no path, redirect to the root path:
        req = trans.request()
        if not requestPath:
            p = req.servletPath() + '/'
            q = req.queryString()
            if q:
                p += "?" + q
            raise HTTPMovedPermanently(location=p)
        # Determine the context name:
        context = [p for p in requestPath.split('/') if p]
        if requestPath.endswith('/'):
            context.append('')
        parts = []
        while context:
            contextName = '/'.join(context)
            if contextName in self._contexts:
                break
            parts.insert(0, context.pop())
        if context:
            if parts:
                parts.insert(0, '')
                requestPath = '/'.join(parts)
            else:
                requestPath = ''
        else:
            contextName = self._defaultContext
        context = self._contexts[contextName]
        req._serverSideContextPath = context
        req._contextName = contextName
        fp = FileParser(context)
        return fp.parse(trans, requestPath)


class _FileParser(URLParser):
    """Parse requests to the filesystem.

    FileParser dispatches to servlets in the filesystem, as well as providing
    hooks to override the FileParser.

    FileParser objects are threadsafe. A factory function is used to cache
    FileParser instances, so for any one path only a single FileParser instance
    will exist.  The `_FileParser` class is the real class, and `FileParser` is
    a factory that either returns an existent _FileParser object, or creates a
    new one if none exists.

    FileParser uses several settings from ``Application.config``, which are
    persistent over the life of the application. These are set up in the
    function `initApp`, as class variables. They cannot be set when the module
    is loaded, because the Application is not yet set up, so `initApp` is
    called in `Application.__init__`.
    """

    # compiled modules with these extensions will be loaded
    # even if no source file exists
    _byteCodeExtensions = ['opt-2.pyc']

    # endregion Parsing

    # region Init

    def __init__(self, path):
        """Create a FileParser.

        Each parsed directory has a FileParser instance associated with it
        (``self._path``).
        """
        URLParser.__init__(self)
        self._path = path
        self._initModule = None

    # endregion Init

    # region Parsing

    def parse(self, trans, requestPath):
        """Return the servlet.

        __init__ files will be used for various hooks
        (see `parseInit` for more).

        If the next part of the URL is a directory, it calls
        ``FileParser(dirPath).parse(trans, restOfPath)`` where ``restOfPath``
        is `requestPath` with the first section of the path removed (the part
        of the path that this FileParser just handled).

        This uses `fileNamesForBaseName` to find files in its directory.
        That function has several functions to define what files are ignored,
        hidden, etc.  See its documentation for more information.
        """
        if debug:
            print(f"FP({self._path!r}) parses {requestPath!r}")

        req = trans.request()

        # First decode the URL, since we are dealing with filenames here:
        # We must use unquote instead of unquote_plus, because decoding
        # the plus sign should only happen for the query string.
        requestPath = unquote(requestPath)

        result = self.parseInit(trans, requestPath)
        if result is not None:
            return result

        if not requestPath or requestPath == '/':
            return self.parseIndex(trans, requestPath)

        if not requestPath.startswith('/'):
            raise HTTPNotFound(f"Invalid path info: {requestPath}")

        parts = requestPath[1:].split('/', 1)
        nextPart = parts[0]
        restPart = '/' + parts[1] if len(parts) > 1 else ''

        baseName = os.path.join(self._path, nextPart)
        if restPart and not self._extraPathInfo:
            names = [baseName]
        else:
            names = self.filenamesForBaseName(baseName)

        if len(names) > 1:
            warn("More than one file matches"
                 f"{requestPath} in {self._path}: {names}")
            raise HTTPNotFound("Page is ambiguous")
        if not names:
            return self.parseIndex(trans, requestPath)

        name = names[0]
        if os.path.isdir(name):
            # directories are dispatched to FileParsers
            # rooted in that directory
            fp = FileParser(name)
            return fp.parse(trans, restPart)

        req._extraURLPath = restPart

        if not self._extraPathInfo and restPart:
            raise HTTPNotFound(f"Invalid extra path info: {restPart}")

        req._serverSidePath = name

        return ServletFactoryManager.servletForFile(trans, name)

    def filenamesForBaseName(self, baseName):
        """Find all files for a given base name.

        Given a path, like ``/a/b/c``, searches for files in ``/a/b``
        that start with ``c``.  The final name may include an extension,
        which is less ambiguous; though if you ask for file.html,
        and file.html.py exists, that file will be returned.

        The files are filtered according to the settings ``FilesToHide``,
        ``FilesToServe``, ``ExtensionsToIgnore`` and ``ExtensionsToServe``.
        See the shouldServeFile() method for details on these settings.

        All files that start with the given base name are returned
        as a list. When the base name itself is part of the list or
        when extensions are prioritized and such an extension is found
        in the list, then the list will be reduced to only that entry.

        Some settings are used to control the prioritization of filenames.
        All settings are in ``Application.config``:

        UseCascadingExtensions:
            If true, then extensions will be prioritized.  So if
            extension ``.tmpl`` shows up in ExtensionCascadeOrder
            before ``.html``, then even if filenames with both
            extensions exist, only the .tmpl file will be returned.
        ExtensionCascadeOrder:
            A list of extensions, ordered by priority.
        """
        if '*' in baseName:
            return []

        fileStart = os.path.basename(baseName)
        dirName = os.path.dirname(baseName)
        filenames = []
        for filename in os.listdir(dirName):
            if filename.startswith('.'):
                continue
            if filename == fileStart:
                if self.shouldServeFile(filename):
                    return [os.path.join(dirName, filename)]
            elif (filename.startswith(fileStart)
                  and os.path.splitext(filename)[0] == fileStart):
                if self.shouldServeFile(filename):
                    filenames.append(os.path.join(dirName, filename))

        if self._useCascading and len(filenames) > 1:
            for extension in self._cascadeOrder:
                if baseName + extension in filenames:
                    return [baseName + extension]

        return filenames

    def shouldServeFile(self, filename):
        """Check if the file with the given filename should be served.

        Some settings are used to control the filtering of filenames.
        All settings are in ``Application.config``:

        FilesToHide:
            These files will be ignored, and even given a full
            extension will not be used.  Takes a glob.
        FilesToServe:
            If set, *only* files matching these globs will be
            served, all other files will be ignored.
        ExtensionsToIgnore:
            Files with these extensions will be ignored, but if a
            complete filename (with extension) is given the file
            *will* be served (unlike FilesToHide).  Extensions are
            in the form ``".py"``
        ExtensionsToServe:
            If set, only files with these extensions will be
            served.  Like FilesToServe, only doesn't use globs.
        """
        ext = os.path.splitext(filename)[1]
        if ext in self._toIgnore:
            return False
        if self._toServe and ext not in self._toServe:
            return False
        for regex in self._filesToHideRegexes:
            if regex.match(filename):
                return False
        if self._filesToServeRegexes:
            for regex in self._filesToServeRegexes:
                if regex.match(filename):
                    break
            else:
                return False
        return True

    def parseIndex(self, trans, requestPath):
        """Return index servlet.

        Return the servlet for a directory index (i.e., ``Main`` or
        ``index``).  When `parse` encounters a directory and there's nothing
        left in the URL, or when there is something left and no file matches
        it, then it will try `parseIndex` to see if there's an index file.

        That means that if ``/a/b/c`` is requested, and in ``/a`` there's no
        file or directory matching ``b``, then it'll look for an index file
        (like ``Main.py``), and that servlet will be returned. In fact, if
        no ``a`` was found, and the default context had an index (like
        ``index.html``) then that would be called with ``/a/b/c`` as
        `HTTPRequest.extraURLPath`.  If you don't want that to occur, you
        should raise an HTTPNotFound in your no-extra-url-path-taking servlets.

        The directory names are based off the ``Application.config`` setting
        ``DirectoryFile``, which is a list of base names, by default
        ``["Main", "index", "main", "Index"]``, which are searched in order.
        A file with any extension is allowed, so the index can be an HTML file,
        a PSP file, a Python servlet, etc.
        """
        req = trans.request()
        # If requestPath is empty, then we're missing the trailing slash:
        if not requestPath:
            p = req.serverURL() + '/'
            q = req.queryString()
            if q:
                p += "?" + q
            raise HTTPMovedPermanently(location=p)
        if requestPath == '/':
            requestPath = ''
        for directoryFile in self._directoryFile:
            basename = os.path.join(self._path, directoryFile)
            names = self.filenamesForBaseName(basename)
            if len(names) > 1 and self._useCascading:
                for ext in self._cascadeOrder:
                    if basename + ext in names:
                        names = [basename + ext]
                        break
            if len(names) > 1:
                warn("More than one file matches the index file"
                     f" {directoryFile} in {self._path}: {names}")
                raise HTTPNotFound("Index page is ambiguous")
            if names:
                if requestPath and not self._extraPathInfo:
                    raise HTTPNotFound
                req._serverSidePath = names[0]
                req._extraURLPath = requestPath
                return ServletFactoryManager.servletForFile(trans, names[0])
        raise HTTPNotFound("Index page not found")

    def initModule(self):
        """Get the __init__ module object for this FileParser's directory."""
        path = self._path
        # if this directory is a context, return the context package
        for context, contextDir in self._app.contexts().items():
            if contextDir == path:
                # avoid reloading of the context package
                return sys.modules.get(context)
        try:
            spec = self._imp.findSpec('__init__', path)
            self._imp.moduleFromSpec(spec)
        except (ImportError, TypeError):
            pass

    def parseInit(self, trans, requestPath):
        """Parse the __init__ file.

        Returns the resulting servlet, or None if no __init__ hooks were found.

        Hooks are put in by defining special functions or objects in your
        __init__, with specific names:

        `urlTransactionHook`:
            A function that takes one argument (the transaction).
            The return value from the function is ignored.  You
            can modify the transaction with this function, though.

        `urlRedirect`:
            A dictionary.  Keys in the dictionary are source
            URLs, the value is the path to redirect to, or a
            `URLParser` object to which the transaction should
            be delegated.

            For example, if the URL is ``/a/b/c``, and we've already
            parsed ``/a`` and are looking for ``b/c``, and we fine
            `urlRedirect`` in a.__init__, then we'll look for a key
            ``b`` in the dictionary.  The value will be a directory
            we should continue to (say, ``/destpath/``).  We'll
            then look for ``c`` in ``destpath``.

            If a key '' (empty string) is in the dictionary, then
            if no more specific key is found all requests will
            be redirected to that path.

            Instead of a string giving a path to redirect to, you
            can also give a URLParser object, so that some portions
            of the path are delegated to different parsers.

            If no matching key is found, and there is no '' key,
            then parsing goes on as usual.

        `SubParser`:
            This should be a class object.  It will be instantiated,
            and then `parse` will be called with it, delegating to
            this instance.  When instantiated, it will be passed
            *this* FileParser instance; the parser can use this to
            return control back to the FileParser after doing whatever
            it wants to do.

            You may want to use a line like this to handle the names::

                from ParserX import ParserX as SubParser

        `urlParser`:
            This should be an already instantiated URLParser-like
            object.  `parse(trans, requestPath)` will be called
            on this instance.

        `urlParserHook`:
            Like `urlParser`, except the method
            `parseHook(trans, requestPath, fileParser)` will
            be called, where fileParser is this FileParser instance.

        `urlJoins`:
            Either a single path, or a list of paths.  You can also
            use URLParser objects, like with `urlRedirect`.

            Each of these paths (or parsers) will be tried in
            order.  If it raises HTTPNotFound, then the next path
            will be tried, ending with the current path.

            Paths are relative to the current directory.  If you
            don't want the current directory to be a last resort,
            you can include '.' in the joins.
        """
        if self._initModule is None:
            self._initModule = self.initModule()
        mod = self._initModule

        seen = trans._fileParserInitSeen[self._path]

        if ('urlTransactionHook' not in seen
                and hasattr(mod, 'urlTransactionHook')):
            seen.add('urlTransactionHook')
            mod.urlTransactionHook(trans)

        if 'urlRedirect' not in seen and hasattr(mod, 'urlRedirect'):
            seen.add('urlRedirect')
            try:
                nextPart, restPath = requestPath[1:].split('/', 1)
                restPath = '/' + restPath
            except ValueError:
                nextPart = requestPath[1:]
                restPath = ''
            if nextPart in mod.urlRedirect:
                redirectTo = mod.urlRedirect[nextPart]
                redirectPath = restPath
            elif '' in mod.urlRedirect:
                redirectTo = mod.urlRedirect['']
                redirectPath = restPath
            else:
                redirectTo = None
            if redirectTo:
                if isinstance(redirectTo, str):
                    fp = FileParser(os.path.join(self._path, redirectTo))
                else:
                    fp = redirectTo
                return fp.parse(trans, redirectPath)

        if 'SubParser' not in seen and hasattr(mod, 'SubParser'):
            seen.add('SubParser')
            parser = mod.SubParser(self)
            return parser.parse(trans, requestPath)

        if 'urlParser' not in seen and hasattr(mod, 'urlParser'):
            seen.add('urlParser')
            parser = mod.urlParser
            return parser.parse(trans, requestPath)

        if 'urlParserHook' not in seen and hasattr(mod, 'urlParserHook'):
            seen.add('urlParserHook')
            parser = mod.urlParserHook
            return parser.parseHook(trans, requestPath, self)

        if 'urlJoins' not in seen and hasattr(mod, 'urlJoins'):
            seen.add('urlJoins')
            joinPath = mod.urlJoins
            if isinstance(joinPath, str):
                joinPath = [joinPath]
            for path in joinPath:
                path = os.path.join(self._path, path)
                if isinstance(path, str):
                    parser = FileParser(os.path.join(self._path, path))
                else:
                    parser = path
                try:
                    return parser.parse(trans, requestPath)
                except HTTPNotFound:
                    pass

        return None

    # endregion Parsing


FileParser = ParamFactory(_FileParser)


class URLParameterParser(URLParser):
    """Strips named parameters out of the URL.

    E.g. in ``/path/SID=123/etc`` the ``SID=123`` will be removed from the URL,
    and a field will be set in the request (so long as no field by that name
    already exists -- if a field does exist the variable is thrown away).
    These are put in the place of GET or POST variables.

    It should be put in an __init__, like::

        from URLParser import URLParameterParser
        urlParserHook = URLParameterParser()

    Or (slightly less efficient):

        from URLParser import URLParameterParser as SubParser
    """

    # region Init

    def __init__(self, fileParser=None):
        self._fileParser = fileParser

    # endregion Init

    # region Parsing

    def parse(self, trans, requestPath):
        """Delegates to `parseHook`."""
        return self.parseHook(trans, requestPath, self._fileParser)

    @staticmethod
    def parseHook(trans, requestPath, hook):
        """Munges the path.

        The `hook` is the FileParser object that originally called this --
        we just want to strip stuff out of the URL and then give it back to
        the FileParser instance, which can actually find the servlet.
        """
        parts = requestPath.split('/')
        result = []
        req = trans.request()
        for part in parts:
            if '=' in part:
                name, value = part.split('=', 1)
                if not req.hasField(name):
                    req.setField(name, value)
            else:
                result.append(part)
        return hook.parse(trans, '/'.join(result))

    # endregion Parsing


class ServletFactoryManagerClass:
    """Manage servlet factories.

    This singleton (called `ServletFactoryManager`) collects and manages
    all the servlet factories that are installed.

    See `addServletFactory` for adding new factories, and `servletForFile`
    for getting the factories back.
    """

    # region Init

    def __init__(self):
        self.reset()

    def reset(self):
        self._factories = []
        self._factoryExtensions = {}

    # endregion Init

    # region Manager

    def addServletFactory(self, factory):
        """Add a new servlet factory.

        Servlet factories can add themselves with::

            ServletFactoryManager.addServletFactory(factory)

        The factory must have an `extensions` method, which should
        return a list of extensions that the factory handles (like
        ``['.ht']``).  The special extension ``.*`` will match any
        file if no other factory is found.  See `ServletFactory`
        for more information.
        """

        self._factories.append(factory)
        for ext in factory.extensions():
            if ext in self._factoryExtensions:
                raise KeyError(
                    f"Extension {ext!r} for factory {factory.__name__}"
                    " was already used by factory"
                    f" {self._factoryExtensions[ext].__name__}")
            self._factoryExtensions[ext] = factory

    def factoryForFile(self, path):
        """Get a factory for a filename."""
        ext = os.path.splitext(path)[1]
        if ext in self._factoryExtensions:
            return self._factoryExtensions[ext]
        if '.*' in self._factoryExtensions:
            return self._factoryExtensions['.*']
        raise HTTPNotFound

    def servletForFile(self, trans, path):
        """Get a servlet for a filename and transaction.

        Uses `factoryForFile` to find the factory, which
        creates the servlet.
        """
        factory = self.factoryForFile(path)
        return factory.servletForTransaction(trans)

    # endregion Manager


ServletFactoryManager = ServletFactoryManagerClass()


# region Global Init

def initApp(app):
    """Initialize the application.

    Installs the proper servlet factories, and gets some settings from
    Application.config. Also saves the application in _globalApplication
    for future calls to the application() function.

    This needs to be called before any of the URLParser-derived classes
    are instantiated.
    """
    global _globalApplication
    _globalApplication = app
    from UnknownFileTypeServlet import UnknownFileTypeServletFactory
    from ServletFactory import PythonServletFactory

    ServletFactoryManager.reset()
    for factory in [UnknownFileTypeServletFactory, PythonServletFactory]:
        ServletFactoryManager.addServletFactory(factory(app))

    initParser(app)


def initParser(app):
    """Initialize the FileParser Class."""
    cls = _FileParser
    cls._app = app
    cls._imp = app._imp
    cls._contexts = app.contexts
    cls._filesToHideRegexes = []
    cls._filesToServeRegexes = []
    from fnmatch import translate as fnTranslate
    for pattern in app.setting('FilesToHide'):
        cls._filesToHideRegexes.append(re.compile(fnTranslate(pattern)))
    for pattern in app.setting('FilesToServe'):
        cls._filesToServeRegexes.append(re.compile(fnTranslate(pattern)))
    cls._toIgnore = app.setting('ExtensionsToIgnore')
    cls._toServe = app.setting('ExtensionsToServe')
    cls._useCascading = app.setting('UseCascadingExtensions')
    cls._cascadeOrder = app.setting('ExtensionCascadeOrder')
    cls._directoryFile = app.setting('DirectoryFile')
    cls._extraPathInfo = app.setting('ExtraPathInfo')

# endregion Global Init



from TaskKit.Task import Task


class SessionTask(Task):
    """The session sweeper task."""

    def __init__(self, sessions):
        Task.__init__(self)
        self._sessionstore = sessions

    def run(self):
        if self.proceed():
            self._sessionstore.cleanStaleSessions(self)



"""Exception handling."""

import os
import sys
import poplib
import smtplib
import traceback
from io import StringIO
from random import randint
from time import time, localtime

from email.message import Message
from email.utils import formatdate

from MiscUtils.Funcs import asclocaltime
from WebUtils.HTMLForException import htmlForException
from WebUtils.Funcs import htmlForDict, htmlEncode


class Singleton:
    """A singleton object."""


class ExceptionHandler:
    """Exception handling.

    ExceptionHandler is a utility class for Application that is created
    to handle a particular exception. The object is a one-shot deal.
    After handling an exception, it should be removed.

    At some point, the exception handler sends `writeExceptionReport` to the
    transaction (if present), which in turn sends it to the other transactional
    objects (application, request, response, etc.)  The handler is the single
    argument for this message.

    Classes may find it useful to do things like this::

        exceptionReportAttrs = ['foo', 'bar', 'baz']
        def writeExceptionReport(self, handler):
            handler.writeTitle(self.__class__.__name__)
            handler.writeAttrs(self, self.exceptionReportAttrs)

    The handler write methods that may be useful are:

      * write
      * writeTitle
      * writeDict
      * writeAttrs

    Derived classes must not assume that the error occurred in a
    transaction.  self._tra may be None for exceptions outside
    of transactions.

    **HOW TO CREATE A CUSTOM EXCEPTION HANDLER**

    In the ``__init__.py`` of your context::

        from ExceptionHandler import ExceptionHandler as _ExceptionHandler

        class ExceptionHandler(_ExceptionHandler):

            _hideValuesForFields = _ExceptionHandler._hideValuesForFields + [
                'foo', 'bar']

            def work(self):
                _ExceptionHandler.work(self)
                # do whatever
                # override other methods if you like

        def contextInitialize(app, ctxPath):
            app._exceptionHandlerClass = ExceptionHandler

    You can also control the errors with settings in ``Application.config``.
    """
    # keep these lower case to support case insensitivity:
    _hideValuesForFields = [
        'password', 'passwd', 'pwd',
        'creditcard', 'credit card', 'cc', 'pin', 'tan']
    _hiddenString = '*** hidden ***'
    _addSpace = {
        'PATH': os.pathsep, 'CLASSPATH': os.pathsep,
        'HTTP_ACCEPT': ',', 'HTTP_ACCEPT_CHARSET': ',',
        'HTTP_ACCEPT_ENCODING': ',', 'HTTP_ACCEPT_LANGUAGE': ','}
    _docType = '<!DOCTYPE html>'
    _emailBody = (
        'Webware caught an exception while processing'
        ' a request for "{path}" at {ascTime} (timestamp: {time}).'
        ' The plain text traceback from Python is printed below and'
        ' the full HTML error report from Webware is attached.\n\n')

    # region Init

    def __init__(
            self, application, transaction, excInfo, formatOptions=None):
        """Create an exception handler instance.

        ExceptionHandler instances are created anew for each exception.
        Instantiating ExceptionHandler completes the process --
        the caller need not do anything else.
        """
        # Keep references to the objects
        self._app = application
        self._tra = transaction
        self._exc = excInfo
        if self._tra:
            self._req = self._tra.request()
            self._res = self._tra.response()
        else:
            self._req = self._res = None

        self._formatOptions = formatOptions

        # Make some repairs, if needed. We use the transaction
        # and response to get the error page back out.

        # Cache MaxValueLengthInExceptionReport for speed
        self._maxValueLength = self.setting('MaxValueLengthInExceptionReport')

        # exception occurrence time (overridden by response.endTime())
        self._time = time()

        # Get to work
        self.work()

    # endregion Init

    # region Accessors

    def setting(self, name):
        """Settings are inherited from Application."""
        return self._app.setting(name)

    def servletPathname(self):
        """The full filesystem path for the servlet."""
        try:
            return self._tra.request().serverSidePath()
        except Exception:
            return None

    def basicServletName(self):
        """The base name for the servlet (sans directory)."""
        name = self.servletPathname()
        if name is None:
            return 'unknown'
        return os.path.basename(name)

    # endregion Accessors

    # region Exception Handling

    def work(self):
        """Main error handling method.

        Invoked by ``__init__`` to do the main work. This calls
        `logExceptionToConsole`, then checks settings to see if it should call
        `saveErrorPage` (to save the error to disk) and `emailException`.

        It also sends gives a page from `privateErrorPage` or
        `publicErrorPage` (which one based on `ShowDebugInfoOnErrors`).
        """
        if self._res:
            self._res.recordEndTime()
            self._time = self._res.endTime()

        self.logExceptionToConsole()

        # Write the error page out to the response if available:
        if self._res and (
                not self._res.isCommitted() or
                self._res.header('Content-Type', None) == 'text/html'):
            if not self._res.isCommitted():
                self._res.reset()
                self._res.setStatus(500, "Servlet Error")
            if self.setting('ShowDebugInfoOnErrors') == 1:
                publicErrorPage = self.privateErrorPage()
            else:
                publicErrorPage = self.publicErrorPage()
            self._res.write(publicErrorPage)

            # Add a large block comment; this prevents IE from overriding the
            # page with its own generic error 500 page
            self._res.write(
                '<!-- - - - - - - - - - - - - - - - - - -->\n' * 100)

        privateErrorPage = None
        if self.setting('SaveErrorMessages'):
            privateErrorPage = self.privateErrorPage()
            filename = self.saveErrorPage(privateErrorPage)
        else:
            filename = None

        self.logExceptionToDisk(errorMsgFilename=filename)

        if self.setting('EmailErrors'):
            if privateErrorPage is None:
                privateErrorPage = self.privateErrorPage()
            try:
                self.emailException(privateErrorPage)
            except Exception as e:
                print("Could not send error email:", e)

    def logExceptionToConsole(self, stderr=None):
        """Log an exception.

        Logs the time, servlet name and traceback to the console
        (typically stderr). This usually results in the information
        appearing in console/terminal from which the Application was launched.
        """
        if stderr is None:
            stderr = sys.stderr
        stderr.write(
            f'[{asclocaltime(self._time)}] [error] Webware:'
            f' Error while executing script {self.servletPathname()}\n')
        traceback.print_exc(file=stderr)

    def publicErrorPage(self):
        """Return a public error page.

        Returns a brief error page telling the user that an error has occurred.
        Body of the message comes from ``UserErrorMessage`` setting.
        """
        return '\n'.join((
            docType(), '<html>', '<head>',
            '<title>Error</title>', htStyle(), '</head>',
            '<body style="color:black;background-color:white">',
            htTitle('Error'),
            '<p>{}</p>'.format(self.setting('UserErrorMessage')),
            '</body>', '</html>\n'))

    def privateErrorPage(self):
        """Return a private error page.

        Returns an HTML page intended for the developer with
        useful information such as the traceback.

        Most of the contents are generated in `htmlDebugInfo`.
        """
        return '\n'.join((
            docType(), '<html>', '<head>',
            '<title>Error</title>', htStyle(), '</head>',
            '<body style="color:black;background-color:white">',
            htTitle('Error'),
            '<p>{}</p>'.format(self.setting('UserErrorMessage')),
            self.htmlDebugInfo(), '</body>', '</html>\n'))

    def htmlDebugInfo(self):
        """Return the debug info.

        Return HTML-formatted debugging information on the current exception.
        Calls `writeHTML`, which uses ``self.write(...)`` to add content.
        """
        self._html = []
        self.writeHTML()
        html = ''.join(self._html)
        self._html = None
        return html

    def writeHTML(self):
        """Write the traceback.

        Writes all the parts of the traceback, invoking:
          * `writeTraceback`
          * `writeMiscInfo`
          * `writeTransaction`
          * `writeEnvironment`
          * `writeIds`
          * `writeFancyTraceback`
        """
        self.writeTraceback()
        self.writeMiscInfo()
        self.writeTransaction()
        self.writeEnvironment()
        self.writeIds()
        self.writeFancyTraceback()

    # endregion Exception Handling

    # region Write Methods

    def write(self, s):
        """Output `s` to the body."""
        self._html.append(str(s))

    def writeln(self, s):
        """Output `s` plus a newline."""
        self._html.append(str(s))
        self._html.append('\n')

    def writeTitle(self, s):
        """Output the sub-heading to define a section."""
        self.writeln(htTitle(s))

    def writeDict(self, d, heading=None, encoded=None):
        """Output a table-formatted dictionary."""
        self.writeln(htmlForDict(
            d, addSpace=self._addSpace,
            filterValueCallBack=self.filterDictValue,
            maxValueLength=self._maxValueLength,
            topHeading=heading, isEncoded=encoded))

    def writeAttrs(self, obj, attrNames):
        """Output object attributes.

        Writes the attributes of the object as given by attrNames.
        Tries ``obj._name`` first, followed by ``obj.name()``.
        Is resilient regarding exceptions so as not to spoil the
        exception report.
        """
        attrs = {}
        for name in attrNames:
            value = getattr(obj, '_' + name, Singleton)  # go for attribute
            try:
                if value is Singleton:
                    value = getattr(obj, name, Singleton)  # go for method
                    if value is Singleton:
                        value = '(could not find attribute or method)'
                    else:
                        try:
                            if callable(value):
                                value = value()
                            value = self.repr(value)
                        except Exception as e:
                            value = (
                                '(exception during method call:'
                                f' {e.__class__.__name__}: {e})')
                else:
                    value = self.repr(value)
            except Exception as e:
                value = (
                    '(exception during value processing:'
                    f' {e.__class__.__name__}: {e})')
            attrs[name] = value
        self.writeDict(attrs, ('Attribute', 'Value'), True)

    # endregion Write Methods

    # region Traceback sections

    def writeTraceback(self):
        """Output the traceback.

        Writes the traceback, with most of the work done
        by `WebUtils.HTMLForException.htmlForException`.
        """
        self.writeTitle('Traceback')
        self.write(f'<p><em>{self.servletPathname()}</em></p>')
        self.write(htmlForException(self._exc, self._formatOptions))

    def writeMiscInfo(self):
        """Output misc info.

        Write a couple little pieces of information about the environment.
        """
        self.writeTitle('MiscInfo')
        info = {
            'time': asclocaltime(self._time),
            'filename': self.servletPathname(),
            'os.getcwd()': os.getcwd(),
            'sys.path': sys.path,
            'sys.version': sys.version,
        }
        self.writeDict(info)

    def writeTransaction(self):
        """Output transaction.

        Lets the transaction talk about itself, using
        `Transaction.writeExceptionReport`.
        """
        if self._tra:
            self._tra.writeExceptionReport(self)
        else:
            self.writeTitle("No current Transaction.")

    def writeEnvironment(self):
        """Output environment.

        Writes the environment this is being run in. This is *not* the
        environment that was passed in with the request (holding the CGI
        information) -- it's just the information from the environment
        that the Application is being executed in.
        """
        self.writeTitle('Environment')
        self.writeDict(os.environ)

    def writeIds(self):
        """Output OS identification.

        Prints some values from the OS (like processor ID).
        """
        self.writeTitle('OS Ids')
        self.writeDict(osIdDict(), ('Name', 'Value'))

    def writeFancyTraceback(self):
        """Output a fancy traceback, using CGITraceback."""
        if self.setting('IncludeFancyTraceback'):
            self.writeTitle('Fancy Traceback')
            try:
                from WebUtils.ExpansiveHTMLForException import (
                    expansiveHTMLForException)
                self.write(expansiveHTMLForException(
                    context=self.setting('FancyTracebackContext')))
            except Exception:
                self.write(
                    '<p>Unable to generate a fancy traceback'
                    ' (uncaught exception)!</p>')
                try:
                    self.write(htmlForException(sys.exc_info()))
                except Exception:
                    self.write(
                        '<p>Unable to even generate a normal traceback'
                        ' of the exception in fancy traceback!</p>')

    def saveErrorPage(self, html):
        """Save the error page.

        Saves the given HTML error page for later viewing by
        the developer, and returns the filename used.
        """
        filename = os.path.join(
            self._app._errorMessagesDir, self.errorPageFilename())
        try:
            with open(filename, 'w') as f:
                f.write(html)
        except IOError:
            sys.stderr.write(
                f'[{asclocaltime(self._time)}] [error] Webware:'
                f' Cannot save error page ({filename})\n')
        else:
            return filename

    def errorPageFilename(self):
        """Create filename for error page.

        Construct a filename for an HTML error page, not including the
        ``ErrorMessagesDir`` setting (which `saveError` adds on).
        """
        # Note: Using the timestamp and a random number is a poor technique
        # for filename uniqueness, but it is fast and good enough in practice.
        name = self.basicServletName()
        ts = '{:4d}-{:02d}-{:02d}-{:02d}-{:02d}-{:02d}'.format(
            *localtime(self._time)[:6])
        rnd = randint(0, 999999)
        return f'Error-{name}-{ts}-{rnd:06d}.html'

    def logExceptionToDisk(self, errorMsgFilename=None):
        """Log the exception to disk.

        Writes a tuple containing (date-time, filename, pathname,
        exception-name, exception-data, error report filename)
        to the errors file (typically 'Errors.csv') in CSV format.
        Invoked by `handleException`.
        """
        if not self.setting('LogErrors'):
            return
        err, msg = self._exc[:2]
        err, msg = err.__name__, str(msg)

        logLine = (
            asclocaltime(self._time),
            self.basicServletName(), self.servletPathname(),
            err, msg, errorMsgFilename or '')

        def fixElement(element):
            element = str(element)
            if ',' in element or '"' in element:
                element = element.replace('"', '""')
                element = f'"{element}"'
            return element

        logLine = ','.join(map(fixElement, logLine)) + '\n'
        filename = self._app.serverSidePath(self.setting('ErrorLogFilename'))
        try:
            if os.path.exists(filename):
                with open(filename, 'a') as f:
                    f.write(logLine)
            else:
                with open(filename, 'w') as f:
                    f.write(
                        'time,filename,pathname,exception name,'
                        'exception data,error report filename\n')
                    f.write(logLine)
        except Exception as e:
            print('Could not write to error log:', e)

    def emailException(self, htmlErrMsg):
        """Email the exception.

        Send the exception via mail, either as an attachment,
        or as the body of the mail.
        """
        message = Message()

        # Construct the message headers
        headers = self.setting('ErrorEmailHeaders').copy()
        headers['Date'] = formatdate()
        headers['Mime-Version'] = '1.0'
        headers['Subject'] = '{} {}: {}'.format(
            headers.get('Subject', '[Webware Error]'), *sys.exc_info()[:2])
        for header, value in headers.items():
            if isinstance(value, (list, tuple)):
                value = ','.join(value)
            message.add_header(header, value)

        # Construct the message body
        if self.setting('EmailErrorReportAsAttachment'):
            # start off with a text/plain part
            text = self._emailBody.format(
                path=self.servletPathname(),
                ascTime=asclocaltime(self._time), time=self._time)
            message.set_type('multipart/mixed')
            part = Message()
            part.set_type('text/plain')
            body = StringIO()
            body.write(text)
            traceback.print_exc(file=body)
            part.set_payload(body.getvalue())
            body.close()
            message.attach(part)
            part = Message()
            # now add htmlErrMsg
            part.add_header('Content-Transfer-Encoding', '7bit')
            part.add_header(
                'Content-Description',
                'HTML version of Webware error message')
            part.add_header(
                'Content-Disposition',
                'attachment', filename='WebwareErrorMsg.html')
            part.set_type('text/html')
            part.set_payload(htmlErrMsg)
            message.attach(part)
        else:
            message.set_type('text/html')
            message.set_payload(htmlErrMsg, 'us-ascii')

        # Send the message
        server = self.setting('ErrorEmailServer')
        # This setting can be: server, server:port, server:port:user:password
        # or server:port:user:password:popserver:popport for "smtp after pop".
        parts = server.split(':', 5)
        server = port = user = passwd = None
        popserver = popssl = popport = None
        try:
            # fetch individual parts until we get an IndexError
            server = parts[0]
            try:
                port = int(parts[1])
            except ValueError:
                pass
            user = parts[2]
            passwd = parts[3]
            popserver = parts[4]
            try:
                popport = int(parts[5])
            except ValueError:
                popport = None
            if parts[6].lower() == 'ssl':
                popssl = True
        except IndexError:
            pass
        if user and passwd and popserver:
            # SMTP after POP
            if popssl is None and popport == 995:
                popssl = True
            popssl = poplib.POP3_SSL if popssl else poplib.POP3
            if popport:
                popserver = popssl(popserver, popport)
            else:
                popserver = popssl(popserver)
            popserver.set_debuglevel(0)
            popserver.user(user)
            popserver.pass_(passwd)
            try:
                popserver.quit()
            except Exception:
                pass
        if port:
            server = smtplib.SMTP(server, port)
        else:
            server = smtplib.SMTP(server)
        try:
            server.set_debuglevel(0)
            if user and passwd and not popserver:
                # SMTP-AUTH
                server.ehlo()
                if server.has_extn('starttls'):
                    server.starttls()
                    server.ehlo()
                server.login(user, passwd)
            body = message.as_string()
            server.sendmail(headers['From'], headers['To'], body)
        finally:
            try:
                server.quit()
            except Exception:
                pass

    # endregion Traceback sections

    # region Filtering

    def filterDictValue(self, value, key, _dict):
        """Filter dictionary values.

        Filters keys from a dict.  Currently ignores the
        dictionary, and just filters based on the key.
        """
        return self.filterValue(value, key)

    def filterValue(self, value, key):
        """Filter values.

        This is the core filter method that is used in all filtering.
        By default, it simply returns self._hiddenString if the key is
        in self._hideValuesForField (case insensitive). Subclasses
        could override for more elaborate filtering techniques.
        """
        try:
            key = key.lower()
        except Exception:
            pass
        if key in self._hideValuesForFields:
            return self._hiddenString
        return value

    # endregion Filtering

    # region Utility

    def repr(self, value):
        """Get HTML encoded representation.

        Returns the repr() of value already HTML encoded. As a special case,
        dictionaries are nicely formatted in table.

        This is a utility method for `writeAttrs`.
        """
        if isinstance(value, dict):
            return htmlForDict(
                value, addSpace=self._addSpace,
                filterValueCallBack=self.filterDictValue,
                maxValueLength=self._maxValueLength)
        rep = repr(value)
        if self._maxValueLength and len(rep) > self._maxValueLength:
            rep = rep[:self._maxValueLength] + '...'
        return htmlEncode(rep)

    # endregion Utility


# region Misc functions

def docType():
    """Return the document type for the page."""
    return '<!DOCTYPE html>'


def htStyle():
    """Return the page style."""
    return '''<style>
body {
    background-color: white;
    color: #080810;
    font-size: 11pt;
    font-family: Tahoma,Verdana,Arial,Helvetica,sans-serif;
    margin: 0pt;
    padding: 8pt;
}
h2.section {
    font-size: 14pt;
    background-color:#933;
    color:white;
    text-align:center;
}
table.NiceTable {
    font-size: 10pt;
}
table.NiceTable td {
    background-color: #eee;
    color: #111;
}
table.NiceTable th {
    background-color: #bbb;
    color: black;
    text-align: left;
}
table.NiceTable tr.TopHeading th {
    background-color: #555;
    color: white;
    text-align: center;
}
table.NiceTable table.NiceTable td {
    background-color: #ddd;
    color: #222;
}
table.NiceTable table.NiceTable th {
    background-color: #ccc;
    color: black;
    font-weight: normal;
}
</style>'''


def htTitle(name):
    """Format a `name` as a section title."""
    return f'<h2 class="section">{name}</h2>'


def osIdDict():
    """Get all OS id information.

    Returns a dictionary containing id information such as pid and uid.
    """
    ids = ('egid', 'euid', 'gid', 'groups', 'pgrp', 'pid', 'ppid', 'uid')
    attrs = {}
    for id_ in ids:
        try:
            value = getattr(os, 'get' + id_)()
        except AttributeError:
            pass
        else:
            attrs[id_] = value
    return attrs

# endregion Misc functions



try:
    import dominate
except ImportError:
    dominate = None
else:
    from dominate.tags import a, div, h2, p, table, tbody, td, th, tr
    from dominate.util import text

from .ExamplePage import ExamplePage


class DominateDemo(ExamplePage):
    """Demo for using the Dominate library."""

    homepage = 'https://github.com/Knio/dominate'

    def writeContent(self):
        self.writeln('<h1>Using Webware with Dominate</h1>')
        self.writeln(
            '<p>Dominate is a Python library that can be used in Webware'
            ' applications to generate HTML programmatically.</p>')
        if not dominate:
            self.writeln(
                f'<p>Please install <a href="{self.homepage}">Dominate</a>'
                ' in order to view this demo.</p>')
            return

        content = div(id='content')
        with content:
            h2('Hello World!')
            with table(cls="NiceTable").add(tbody()):
                tr(th('Demo table', colspan=3))
                r = tr()
                r += td('One')
                r.add(td('Two'))
                with r:
                    td('Three')
            para = p(__pretty=False)
            with para:
                text('This content has been produced with ')
                a('Dominate', href=self.homepage)
                text(' programmatically.')
        self.write(content)



import os

from SessionMemoryStore import SessionMemoryStore

from .Session import Session
from .TestSessionStore import TestSessionStore


class TestSessionMemoryStore(TestSessionStore):

    _storeClass = SessionMemoryStore
    _storeIsOrdered = True

    def setUp(self):
        d = self._app._sessionDir
        if not os.path.exists(d):
            os.mkdir(d)
        TestSessionStore.setUp(self)
        self._store.clear()
        for n in range(7):
            session = Session(n)
            self._store[session.identifier()] = session
            self.assertFalse(session.isExpired())

    def tearDown(self):
        self._store.clear()
        self._store.storeAllSessions()
        d = self._app._sessionDir
        for filename in os.listdir(d):
            os.remove(os.path.join(d, filename))
        os.rmdir(d)
        TestSessionStore.tearDown(self)

    def testLen(self):
        self.assertEqual(len(self._store), 7)

    def testGetItem(self):
        store = self._store
        self.assertEqual(self._store['foo-3'].bar(), 18)
        self.assertRaises(KeyError, store.__getitem__, 'foo-7')

    def testSetItem(self):
        session = Session()
        key = session.identifier()
        self._store[key] = session
        self.assertEqual(self._store[key].data(), session.data())

    def testDelItem(self):
        del self._store['foo-3']
        self.assertFalse('foo-3' in self._store)
        self.assertEqual(Session._lastExpired, 'foo-3')
        try:
            self._store['foo-3']
        except KeyError:
            pass
        self.assertRaises(KeyError, self._store.__delitem__, 'foo-3')

    def testContains(self):
        store = self._store
        self.assertTrue(
            'foo-0' in store and 'foo-3' in store and 'foo-6' in store)
        self.assertFalse('foo' in store)
        self.assertFalse('0' in store or '3' in store or '6' in store)
        self.assertFalse('foo-7' in store)

    def testIter(self):
        keys = list(iter(self._store))
        expectedKeys = [f'foo-{n}' for n in range(7)]
        if not self._storeIsOrdered:
            keys = set(keys)
            expectedKeys = set(expectedKeys)
        self.assertEqual(keys, expectedKeys)

    def testKeys(self):
        keys = self._store.keys()  # pylint: disable=assignment-from-no-return
        self.assertTrue(isinstance(keys, list))
        expectedKeys = [f'foo-{n}' for n in range(7)]
        if not self._storeIsOrdered:
            keys = set(keys)
            expectedKeys = set(expectedKeys)
        self.assertEqual(keys, expectedKeys)

    def testClear(self):
        store = self._store
        store.clear()
        self.assertFalse(
            'foo-0' in store or 'foo-3' in store or 'foo-6' in store)

    def testSetDefault(self):
        store = self._store
        session = Session()
        self.assertEqual(store.setdefault('foo-3').bar(), 18)
        self.assertEqual(store.setdefault('foo-3', session).bar(), 18)
        self.assertEqual(store.setdefault('foo-7', session).bar(), 42)
        self.assertEqual(store.setdefault('foo-7').bar(), 42)

    def testPop(self):
        store = self._store
        session = self._store['foo-3']
        self.assertFalse(session.isExpired())
        self.assertEqual(store.pop('foo-3').bar(), 18)
        self.assertRaises(KeyError, store.pop, 'foo-3')
        self.assertEqual(store.pop('foo-3', Session()).bar(), 42)
        self.assertRaises(KeyError, store.pop, 'foo-3')
        self.assertFalse(session.isExpired())

    def testGet(self):
        self.assertEqual(self._store.get('foo-4').bar(), 24)
        self.assertEqual(self._store.get('foo-4', Session()).bar(), 24)
        self.assertEqual(self._store.get('foo-7'), None)
        self.assertEqual(self._store.get('foo-7', Session()).bar(), 42)

    def testStoreSession(self):
        session = self._store['foo-3']
        self.assertEqual(session.bar(), 18)
        session.setBar(19)
        self.assertEqual(session.bar(), 19)
        self._store.storeSession(session)
        session = self._store['foo-3']
        self.assertEqual(session.bar(), 19)
        session = Session(3, 20)
        self._store.storeSession(session)
        session = self._store['foo-3']
        self.assertEqual(session.bar(), 20)
        session = Session()
        self._store.storeSession(session)
        session = self._store['foo-7']
        self.assertEqual(session.bar(), 42)

    def testItems(self):
        items = self._store.items()
        self.assertTrue(isinstance(items, list))
        self.assertEqual(len(items), 7)
        self.assertTrue(isinstance(items[4], tuple))
        self.assertEqual(len(items[4]), 2)
        self.assertEqual(dict(items)['foo-3'].bar(), 18)

    def testIterItems(self):
        items = iter(self._store.items())
        self.assertFalse(isinstance(items, list))
        items = list(items)
        self.assertTrue(isinstance(items[4], tuple))
        self.assertEqual(len(items[4]), 2)
        self.assertEqual(dict(items)['foo-3'].bar(), 18)

    def testValues(self):
        values = self._store.values()
        self.assertTrue(isinstance(values, list))
        self.assertEqual(len(values), 7)
        value = values[4]
        self.assertTrue(isinstance(value, Session))
        self.assertEqual(self._store[value.identifier()].bar(), value.bar())

    def testIterValues(self):
        values = iter(self._store.values())
        self.assertFalse(isinstance(values, list))
        values = list(values)
        self.assertEqual(len(values), 7)
        value = values[4]
        self.assertTrue(isinstance(value, Session))
        self.assertEqual(self._store[value.identifier()].bar(), value.bar())

    def testCleanStaleSessions(self):
        store = self._store
        self.assertEqual(len(store), 7)
        self.assertTrue('foo-0' in store and 'foo-4' in store)
        self.assertTrue('foo-5' in store and 'foo-6' in store)
        store.cleanStaleSessions()
        self.assertEqual(len(store), 5)
        self.assertTrue('foo-0' in store and 'foo-4' in store)
        self.assertFalse('foo-5' in store or 'foo-6' in store)



name = 'MiscUtils'

status = 'stable'

synopsis = (
    "MiscUtils provides support classes and functions to Webware"
    " that aren't necessarily web-related and that don't fit into one of"
    " the other plug-ins. There is plenty of useful reusable code here.")



"""Automated tests for PSPCompiler

Copyright (c) by Winston Wolff, 2004 https://www.stratolab.com
"""

import os
import importlib
import shutil
import sys
import tempfile
import unittest

from io import StringIO
from time import sleep

from PSP import Context, PSPCompiler


class TestCompiler(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testDir = tempfile.mkdtemp()
        sys.path.insert(0, cls.testDir)

    @classmethod
    def tearDownClass(cls):
        sys.path.remove(cls.testDir)
        shutil.rmtree(cls.testDir)

    @staticmethod
    def sync():
        """Make sure everything is written to disk."""
        try:
            os.sync()
        except AttributeError:
            pass  # not Unix
        sleep(0.05)

    def compileString(self, pspSource, classname):
        """Compile a string to an object.

        Takes a string, compiles it, imports the Python file, and returns you
        the Python class object.

        classname = some string so that each file is unique per test case.
        """
        # write string to temporary file
        moduleName = "tmp_TestCompiler_" + classname
        modulePath = os.path.join(self.testDir, moduleName)
        tmpInName = modulePath + '.psp'
        tmpOutName = modulePath + '.py'
        with open(tmpInName, 'w') as fp:
            fp.write(pspSource)
        # Compile PSP into .py file
        context = Context.PSPCLContext(tmpInName)
        context.setClassName(classname)
        context.setPythonFileName(tmpOutName)
        clc = PSPCompiler.Compiler(context)
        clc.compile()
        self.sync()
        self.assertTrue(os.path.isfile(tmpOutName))
        # Have Python import the .py file.
        theModule = importlib.__import__(moduleName)
        # want to return the class inside the module.
        theClass = getattr(theModule, classname)
        return theClass

    def compileAndRun(self, pspSource, classname):
        pspClass = self.compileString(pspSource, classname)
        pspInstance = pspClass()
        outStream = StringIO()
        pspInstance._writeHTML(outStream)
        output = outStream.getvalue()
        return output

    def testExpression(self):
        output = self.compileAndRun(
            'two plus three is: <%= 2+3 %>', 'testExpression')
        self.assertEqual("two plus three is: 5", output)

    def testScript(self):
        output = self.compileAndRun(
            'one plus two is: <% res.write(str(1+2)) %>', 'testScript')
        self.assertEqual("one plus two is: 3", output)

    def testScriptNewLines(self):
        psp = '''\nthree plus two is: \n<%\nres.write(str(3+2)) \n%>'''
        expect = '''\nthree plus two is: \n5'''
        output = self.compileAndRun(psp, 'testScriptNewLines')
        self.assertEqual(output, expect)
        psp = '''\nthree plus two is: \n<%\n  res.write(str(3+2)) \n%>'''
        expect = '''\nthree plus two is: \n5'''
        output = self.compileAndRun(psp, 'testScriptNewLines')
        self.assertEqual(output, expect)

    def testScriptReturns(self):
        psp = '''\rthree plus two is: \r<%\rres.write(str(3+2)) \r%>'''
        expect = '''\nthree plus two is: \n5'''
        output = self.compileAndRun(psp, 'testScriptReturns')
        self.assertEqual(output, expect)
        psp = '''\rthree plus two is: \r<%\r     res.write(str(3+2)) \r%>'''
        expect = '''\nthree plus two is: \n5'''
        output = self.compileAndRun(psp, 'testScriptReturns')
        self.assertEqual(output, expect)

    def testScriptIf(self):
        psp = '''PSP is <% if 1: %>Good<% end %>'''
        expect = '''PSP is Good'''
        output = self.compileAndRun(psp, 'testScriptIf')
        self.assertEqual(output, expect)

    def testScriptIfElse(self):
        psp = '''JSP is <% if 0: %>Good<% end %><% else: %>Bad<% end %>'''
        expect = '''JSP is Bad'''
        output = self.compileAndRun(psp, 'testScriptIfElse')
        self.assertEqual(output, expect)

    def testScriptBlocks(self):
        psp = '''
<% for i in range(3): %>
<%= i %><br/>
<% end %>'''
        expect = '''

0<br/>

1<br/>

2<br/>
'''
        output = self.compileAndRun(psp, 'testScriptBlocks')
        self.assertEqual(output, expect)

    def testScriptBraces(self):
        psp = '''\
<%@page indentType="braces" %>
<% for i in range(3): { %>
<%= i %><br/>
<% } %>'''
        expect = '''

0<br/>

1<br/>

2<br/>
'''
        output = self.compileAndRun(psp, 'testScriptBraces')
        self.assertEqual(output, expect)

    def testPspMethod(self):
        psp = '''
        <psp:method name="add" params="a,b">
        return a+b
        </psp:method>
        7 plus 8 = <%= self.add(7,8) %>
        '''
        output = self.compileAndRun(psp, 'testPspMethod').strip()
        self.assertEqual("7 plus 8 = 15", output)

    def testPspFile(self):
        psp = '''
        <psp:file>
            def square(a):
                return a*a
        </psp:file>
        7^2 = <%= square(7) %>
        '''
        output = self.compileAndRun(psp, 'testPspFile').strip()
        self.assertEqual("7^2 = 49", output)

    def testPspClass(self):
        psp = '''
        <psp:class>
            def square(self, a):
                return a*a
        </psp:class>
        4^2 = <%= self.square(4) %>
        '''
        output = self.compileAndRun(psp, 'testPspClass').strip()
        self.assertEqual("4^2 = 16", output)



from os import sep

from WebUtils.Funcs import urlEncode

from .DumpCSV import DumpCSV


class Errors(DumpCSV):

    def filename(self):
        return self.application().setting('ErrorLogFilename')

    def cellContents(self, _rowIndex, colIndex, value):
        """Hook for subclasses to customize the contents of a cell.

        Based on any criteria (including location).
        """
        if self._headings[colIndex] in ('pathname', 'error report filename'):
            path = self.application().serverSidePath()
            if value.startswith(path):
                value = value[len(path):]
                if value.startswith(sep):
                    value = value[len(sep):]
                link = f'View?filename={urlEncode(value)}'
                value = value.replace(sep, sep + '<wbr>')
                value = f'<a href="{link}">{value}</a>'
            else:
                value = value.replace(sep, sep + '<wbr>')
            return value
        if self._headings[colIndex] == 'time':
            return f'<span style="white-space:nowrap">{value}</span>'
        return self.htmlEncode(value)



"""DataTable.py


INTRODUCTION

This class is useful for representing a table of data arranged by named
columns, where each row in the table can be thought of as a record::

    name   phoneNumber
    ------ -----------
    Chuck  893-3498
    Bill   893-0439
    John   893-5901

This data often comes from delimited text files which typically
have well defined columns or fields with several rows each of which can
be thought of as a record.

Using a DataTable can be as easy as using lists and dictionaries::

    table = DataTable('users.csv')
    for row in table:
        print(row['name'], row['phoneNumber'])

Or even::

    table = DataTable('users.csv')
    for row in table:
        print('{name} {phoneNumber}'.format(**row))

The above print statement relies on the fact that rows can be treated
like dictionaries, using the column headings as keys.

You can also treat a row like an array::

    table = DataTable('something.tabbed', delimiter='\t')
    for row in table:
        for item in row:
            print(item, end=' ')
        print()


COLUMNS

Column headings can have a type specification like so::

    name, age:int, zip:int

Possible types include string, int, float and datetime.

String is assumed if no type is specified but you can set that
assumption when you create the table::

    table = DataTable(headings, defaultType='float')

Using types like int and float will cause DataTable to actually
convert the string values (perhaps read from a file) to these types
so that you can use them in natural operations. For example::

    if row['age'] > 120:
        self.flagData(row, 'age looks high')

As you can see, each row can be accessed as a dictionary with keys
according the column headings. Names are case sensitive.


ADDING ROWS

Like Python lists, data tables have an append() method. You can append
TableRecords, or you pass a dictionary, list or object, in which case a
TableRecord is created based on given values. See the method docs below
for more details.


FILES

By default, the files that DataTable reads from are expected to be
comma-separated value files.

Limited comments are supported: A comment is any line whose very first
character is a #. This allows you to easily comment out lines in your
data files without having to remove them.

Whitespace around field values is stripped.

You can control all this behavior through the arguments found in the
initializer and the various readFoo() methods::

    ...delimiter=',', allowComments=True, stripWhite=True

For example::

    table = DataTable('foo.tabbed', delimiter='\t',
        allowComments=False, stripWhite=False)

You should access these parameters by their name since additional ones
could appear in the future, thereby changing the order.

If you are creating these text files, we recommend the
comma-separated-value format, or CSV. This format is better defined
than the tab delimited format, and can easily be edited and manipulated
by popular spreadsheets and databases.


MICROSOFT EXCEL

On Microsoft Windows systems with Excel and the PyWin32 package
(https://github.com/mhammond/pywin32), DataTable will use Excel
(via COM) to read ".xls" files::

    from MiscUtils import DataTable
    assert DataTable.canReadExcel()
    table = DataTable.DataTable('foo.xls')

With consistency to its CSV processing, DataTable will ignore any row
whose first cell is '#' and strip surrounding whitespace around strings.


TABLES FROM SCRATCH

Here's an example that constructs a table from scratch::

    table = DataTable(['name', 'age:int'])
    table.append(['John', 80])
    table.append({'name': 'John', 'age': 80})
    print(table)


QUERIES

A simple query mechanism is supported for equality of fields::

    matches = table.recordsEqualTo({'uid': 5})
    if matches:
        for match in matches:
            print(match)
    else:
        print('No matches.')


COMMON USES

  * Programs can keep configuration and other data in simple comma-
    separated text files and use DataTable to access them. For example, a
    web site could read its sidebar links from such a file, thereby allowing
    people who don't know Python (or even HTML) to edit these links without
    having to understand other implementation parts of the site.

  * Servers can use DataTable to read and write log files.


FROM THE COMMAND LINE

The only purpose in invoking DataTable from the command line is to see
if it will read a file::

  > python DataTable.py foo.csv

The data table is printed to stdout.


CACHING

DataTable uses "pickle caching" so that it can read .csv files faster
on subsequent loads. You can disable this across the board with::

    from MiscUtils.DataTable import DataTable
    DataTable._usePickleCache = False

Or per instance by passing "usePickleCache=False" to the constructor.

See the docstring of PickleCache.py for more information.


MORE DOCS

Some of the methods in this module have worthwhile doc strings to look at.
See below.


TO DO

  * Allow callback parameter or setting for parsing CSV records.
  * Perhaps TableRecord should inherit list and dict and override
    methods as appropriate?
  * _types and _blankValues aren't really packaged, advertised or
    documented for customization by the user of this module.
  * DataTable:
      * Parameterize the TextColumn class.
      * Parameterize the TableRecord class.
      * More list-like methods such as insert()
      * writeFileNamed() is flawed: it doesn't write the table column type
      * Should it inherit from list?
  * Add error checking that a column name is not a number (which could
    cause problems).
  * Reading Excel sheets with xlrd, not only with win32com.
"""


import os
import sys

from datetime import date, datetime, time, timedelta, tzinfo
from decimal import Decimal
from warnings import warn

from MiscUtils import NoDefault

from .CSVParser import CSVParser
from .CSVJoiner import joinCSVFields
from .Funcs import positiveId

# region Types and blank Values

_types = {
    'str': str,
    'string': str,
    'unicode': str,
    'basestring': str,
    'bytes': bytes,
    'bytestring': bytes,
    'int': int,
    'bool': bool,
    'long': int,
    'decimal': Decimal,
    'float': float,
    'date': date,
    'datetime': datetime,
    'time': time,
    'timedelta': timedelta,
    'tzinfo': tzinfo
}

_blankValues = {
    str: '',
    str: '',
    str: '',
    bool: False,
    int: 0,
    int: 0,
    float: 0.0,
    Decimal: Decimal('0')
}

# endregion Types and blank Values


# region Functions

def canReadExcel():
    try:
        from win32com.client import Dispatch  # pylint: disable=import-error
        Dispatch("Excel.Application")
    except Exception:
        return False
    else:
        return True

# endregion Functions


# region Classes

class DataTableError(Exception):
    """Data table error."""


class TableColumn:
    """Representation of a table column.

    A table column represents a column of the table including name and type.
    It does not contain the actual values of the column. These are stored
    individually in the rows.
    """

    # region Basics

    def __init__(self, spec):
        """Initialize the table column.

        The spec parameter is a string such as 'name' or 'name:type'.
        """
        name, type_ = spec.split(':', 1) if ':' in spec else (spec, None)
        self._name = name
        self.setType(type_)

    def name(self):
        return self._name

    def type(self):
        return self._type

    def setType(self, type_):
        """Set the type (by a string containing the name) of the heading.

        Usually invoked by DataTable to set the default type for columns
        whose types were not specified.
        """
        if type_:
            try:
                self._type = _types[type_.lower()]
            except Exception:
                types = ', '.join(sorted(_types))
                raise DataTableError(f'Unknown type {type_!r}.'
                                     f' Known types are: {types}') from None
        else:
            self._type = None

    def __repr__(self):
        return (f'<{self.__class__.__name__} {self._name}'
                f' with {self._type} at {positiveId(self):x}>')

    def __str__(self):
        return self._name

    # endregion Basics

    # region Utilities

    def valueForRawValue(self, value):
        """Set correct type for raw value.

        The rawValue is typically a string or value already of the appropriate
        type. TableRecord invokes this method to ensure that values (especially
        strings that come from files) are the correct types (e.g., ints are
        ints and floats are floats).
        """
        if self._type:
            if isinstance(value, str) and self._type is bytes:
                return value.encode('utf-8')
            if isinstance(value, bytes) and self._type is str:
                try:
                    return value.decode('utf-8')
                except UnicodeDecodeError:
                    return value.decode('latin-1')
            if value == '' and self._type in (int, int, float, Decimal):
                value = '0'
            if not isinstance(value, self._type):
                value = self._type(value)
        return value

    # endregion Utilities


class DataTable:
    """Representation of a data table.

    See the doc string for this module.
    """

    _usePickleCache = True

    # region Init

    def __init__(
            self, filenameOrHeadings=None, delimiter=',',
            allowComments=True, stripWhite=True, encoding=None,
            defaultType=None, usePickleCache=None):
        if usePickleCache is None:
            self._usePickleCache = self._usePickleCache
        else:
            self._usePickleCache = usePickleCache
        if defaultType and defaultType not in _types:
            raise DataTableError(
                f'Unknown type for default type: {defaultType!r}')
        self._defaultType = defaultType
        self._filename = None
        self._headings = []
        self._rows = []
        if filenameOrHeadings:
            if isinstance(filenameOrHeadings, str):
                self.readFileNamed(
                    filenameOrHeadings, delimiter,
                    allowComments, stripWhite, encoding)
            else:
                self.setHeadings(filenameOrHeadings)

    # endregion Init

    # region File I/O

    def readFileNamed(
            self, filename, delimiter=',',
            allowComments=True, stripWhite=True, encoding=None,
            worksheet=1, row=1, column=1):
        self._filename = filename
        data = None
        if self._usePickleCache:
            from .PickleCache import readPickleCache, writePickleCache
            data = readPickleCache(filename, source='MiscUtils.DataTable')
        if data is None:
            if self._filename.lower().endswith('.xls'):
                self.readExcel(worksheet, row, column)
            else:
                with open(self._filename, encoding=encoding) as f:
                    self.readFile(f, delimiter, allowComments, stripWhite)
            if self._usePickleCache:
                writePickleCache(self, filename, source='MiscUtils.DataTable')
        else:
            self.__dict__ = data.__dict__
        return self

    def readFile(
            self, file, delimiter=',',
            allowComments=True, stripWhite=True):
        return self.readLines(
            file.readlines(), delimiter, allowComments, stripWhite)

    def readString(
            self, string, delimiter=',',
            allowComments=True, stripWhite=True):
        return self.readLines(
            string.splitlines(), delimiter, allowComments, stripWhite)

    def readLines(
            self, lines, delimiter=',', allowComments=True, stripWhite=True):
        if self._defaultType is None:
            self._defaultType = 'str'
        haveReadHeadings = False
        parse = CSVParser(
            fieldSep=delimiter, allowComments=allowComments,
            stripWhitespace=stripWhite).parse
        for line in lines:
            # process a row, either headings or data
            values = parse(line)
            if values:
                if haveReadHeadings:
                    row = TableRecord(self, values)
                    self._rows.append(row)
                else:
                    self.setHeadings(values)
                    haveReadHeadings = True
        if values is None:
            raise DataTableError("Unfinished multiline record.")
        return self

    @staticmethod
    def canReadExcel():
        return canReadExcel()

    def readExcel(self, worksheet=1, row=1, column=1):
        maxBlankRows = 10
        numRowsToReadPerCall = 20
        from win32com.client import Dispatch  # pylint: disable=import-error
        excel = Dispatch("Excel.Application")
        workbook = excel.Workbooks.Open(os.path.abspath(self._filename))
        try:
            worksheet = workbook.Worksheets(worksheet)
            worksheet.Cells(row, column)
            # determine max column
            numCols = 1
            while True:
                if worksheet.Cells(row, numCols).Value in (None, ''):
                    numCols -= 1
                    break
                numCols += 1
            if numCols <= 0:
                return

            def strip(x):
                try:
                    return x.strip()
                except Exception:
                    return x

            # read rows of data
            maxCol = chr(ord('A') + numCols - 1)
            haveReadHeadings = False
            rowNum = row
            numBlankRows = 0
            valuesBuffer = {}  # keyed by row number
            while True:
                try:
                    # grab a single row
                    values = valuesBuffer[rowNum]
                except KeyError:
                    # woops. read buffer is out of fresh rows
                    maxRow = rowNum + numRowsToReadPerCall - 1
                    valuesRows = worksheet.Range(
                        f'A{rowNum}:{maxCol}{maxRow}').Value
                    valuesBuffer.clear()
                    j = rowNum
                    for valuesRow in valuesRows:
                        valuesBuffer[j] = valuesRow
                        j += 1
                    values = valuesBuffer[rowNum]
                values = [strip(v) for v in values]
                nonEmpty = [v for v in values if v]
                if nonEmpty:
                    if values[0] != '#':
                        if haveReadHeadings:
                            row = TableRecord(self, values)
                            self._rows.append(row)
                        else:
                            self.setHeadings(values)
                            haveReadHeadings = True
                    numBlankRows = 0
                else:
                    numBlankRows += 1
                    if numBlankRows > maxBlankRows:
                        # consider end of spreadsheet
                        break
                rowNum += 1
        finally:
            workbook.Close()

    def save(self):
        self.writeFileNamed(self._filename)

    def writeFileNamed(self, filename):
        with open(filename, 'w') as f:
            self.writeFile(f)

    def writeFile(self, file):
        """Write the table out as a file.

        This doesn't write the column types (like int) back out.

        It's notable that a blank numeric value gets read as zero
        and written out that way. Also, values None are written as blanks.
        """
        # write headings
        file.write(','.join(map(str, self._headings)))
        file.write('\n')

        def valueWritingMapper(item):
            # None gets written as a blank and everything else as a string
            if item is None:
                return ''
            if isinstance(item, str):
                return item
            if isinstance(item, bytes):
                return item.decode('utf-8')
            return str(item)

        # write rows
        for row in self._rows:
            file.write(joinCSVFields(map(valueWritingMapper, row)))
            file.write('\n')

    def commit(self):
        if self._changed:
            self.save()
            self._changed = False

    # endregion File I/O

    # region Headings

    def heading(self, index):
        if isinstance(index, str):
            index = self._nameToIndexMap[index]
        return self._headings[index]

    def hasHeading(self, name):
        return name in self._nameToIndexMap

    def numHeadings(self):
        return len(self._headings)

    def headings(self):
        return self._headings

    def setHeadings(self, headings):
        """Set table headings.

        Headings can be a list of strings (like ['name', 'age:int'])
        or a list of TableColumns or None.
        """
        if not headings:
            self._headings = []
        elif isinstance(headings[0], str):
            self._headings = list(map(TableColumn, headings))
        elif isinstance(headings[0], TableColumn):
            self._headings = list(headings)
        for heading in self._headings:
            if heading.type() is None:
                heading.setType(self._defaultType)
        self.createNameToIndexMap()

    # endregion Headings

    # region Row access (list like)

    def __len__(self):
        return len(self._rows)

    def __getitem__(self, index):
        return self._rows[index]

    def append(self, obj):
        """Append an object to the table.

        If obj is not a TableRecord, then one is created,
        passing the object to initialize the TableRecord.
        Therefore, obj can be a TableRecord, list, dictionary or object.
        See TableRecord for details.
        """
        if not isinstance(obj, TableRecord):
            obj = TableRecord(self, obj)
        self._rows.append(obj)
        self._changed = True

    # endregion Row access (list like)

    # region Queries

    def recordsEqualTo(self, record):
        records = []
        for row in self._rows:
            for key in row:
                if record[key] != row[key]:
                    break
            else:
                records.append(row)
        return records

    # endregion Queries

    # region As a string

    def __repr__(self):
        s = [f'DataTable: {self._filename}\n'
             f'{len(self._rows)} rows\n', ' ' * 5,
             ', '.join(map(str, self._headings)), '\n']
        for i, row in enumerate(self._rows):
            s.append(f'{i:3d}. ')
            s.append(', '.join(map(str, row)))
            s.append('\n')
        return ''.join(s)

    # endregion As a string

    # region As a dictionary

    def dictKeyedBy(self, key):
        """Return a dictionary containing the contents of the table.

        The content is indexed by the particular key. This is useful
        for tables that have a column which represents a unique key
        (such as a name, serial number, etc.).
        """
        return {row[key]: row for row in self}

    # endregion As a dictionary

    # region Misc access

    def filename(self):
        return self._filename

    def nameToIndexMap(self):
        """Speed-up index.

        Table rows keep a reference to this map in order to speed up
        index-by-names (as in row['name']).
        """
        return self._nameToIndexMap

    # endregion Misc access

    # region Self utilities

    def createNameToIndexMap(self):
        """Create speed-up index.

        Invoked by self to create the nameToIndexMap after the table's
        headings have been read/initialized.
        """
        self._nameToIndexMap = {
            heading.name(): i for i, heading in enumerate(self._headings)}

    # endregion Self utilities


class TableRecord:
    """Representation of a table record."""

    # region Init

    def __init__(self, table, values=None, headings=None):
        """Initialize table record.

        Dispatches control to one of the other init methods based on the type
        of values. Values can be one of three things:

            1. A TableRecord
            2. A list
            3. A dictionary
            4. Any object responding to hasValueForKey() and valueForKey().
        """
        self._headings = table.headings() if headings is None else headings
        self._nameToIndexMap = table.nameToIndexMap()
        if values is not None:
            if isinstance(values, (list, tuple)):
                self.initFromSequence(values)
            elif isinstance(values, dict):
                self.initFromDict(values)
            else:
                try:
                    self.initFromObject(values)
                except AttributeError as e:
                    raise DataTableError(
                        f'Unknown type for values {values!r}.') from e

    def initFromSequence(self, values):
        headings = self._headings
        numHeadings = len(headings)
        numValues = len(values)
        if numHeadings < numValues:
            raise DataTableError(
                'There are more values than headings.\n'
                f'headings({numHeadings}, {headings})\n'
                f'values({numValues}, {values})')
        self._values = []
        append = self._values.append
        for i, heading in enumerate(headings):
            if i >= numValues:
                append(_blankValues.get(heading.type()))
            else:
                append(heading.valueForRawValue(values[i]))

    def initFromDict(self, values):
        self._values = []
        append = self._values.append
        for heading in self._headings:
            name = heading.name()
            if name in values:
                append(heading.valueForRawValue(values[name]))
            else:
                append(_blankValues.get(heading.type()))

    def initFromObject(self, obj):
        """Initialize from object.

        The object is expected to response to hasValueForKey(name) and
        valueForKey(name) for each of the headings in the table. It's alright
        if the object returns False for hasValueForKey(). In that case, a
        "blank" value is assumed (such as zero or an empty string). If
        hasValueForKey() returns True, then valueForKey() must return a value.
        """
        self._values = []
        append = self._values.append
        for heading in self._headings:
            name = heading.name()
            if obj.hasValueForKey(name):
                append(heading.valueForRawValue(obj.valueForKey(name)))
            else:
                append(_blankValues.get(heading.type()))

    # endregion Init

    # region Accessing like a sequence or dictionary

    def __len__(self):
        return len(self._values)

    def __getitem__(self, key):
        if isinstance(key, str):
            key = self._nameToIndexMap[key]
        try:
            return self._values[key]
        except TypeError:
            raise TypeError(f'key={key!r}, key type={type(key)!r},'
                            f' values={self._values!r}') from None

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key = self._nameToIndexMap[key]
        self._values[key] = value

    def __delitem__(self, key):
        if isinstance(key, str):
            key = self._nameToIndexMap[key]
        del self._values[key]

    def __contains__(self, key):
        return key in self._nameToIndexMap

    def __repr__(self):
        return repr(self._values)

    def __iter__(self):
        for value in self._values:
            yield value

    def get(self, key, default=None):
        index = self._nameToIndexMap.get(key)
        if index is None:
            return default
        return self._values[index]

    def has_key(self, key):
        warn("has_key is deprecated, please us 'in' instead.",
             DeprecationWarning, stacklevel=2)
        return key in self

    def keys(self):
        return list(self._nameToIndexMap.keys())

    def values(self):
        return self._values

    def items(self):
        items = []
        for key in self._nameToIndexMap:
            items.append((key, self[key]))
        return items

    def iterkeys(self):
        return iter(self._nameToIndexMap)

    def itervalues(self):
        return iter(self)

    def iteritems(self):
        for key in self._nameToIndexMap:
            yield key, self[key]

    # endregion Accessing like a sequence or dictionary

    # region Additional access

    def asList(self):
        """Return a sequence whose values are the same as the record's.

        The order of the sequence is the one defined by the table.
        """
        # It just so happens that our implementation already has this
        return self._values[:]

    def asDict(self):
        """Return a dictionary whose key-values match the table record."""
        record = {}
        nameToIndexMap = self._nameToIndexMap
        for key in nameToIndexMap:
            record[key] = self._values[nameToIndexMap[key]]
        return record

    # endregion Additional access

    # region valueFor family

    def valueForKey(self, key, default=NoDefault):
        if default is NoDefault:
            return self[key]
        return self.get(key, default)

    def valueForAttr(self, attr, default=NoDefault):
        return self.valueForKey(attr['Name'], default)

    # endregion valueFor family

# endregion Classes


# region Standalone

def main(args=None):
    if args is None:
        args = sys.argv
    for arg in args[1:]:
        dataTable = DataTable(arg)
        print(f'*** {arg} ***')
        print(dataTable)
        print()


if __name__ == '__main__':
    main()

# endregion Standalone



"""Utility class for keeping track of the context.

A utility class that holds information about the file we are parsing
and the environment we are doing it in.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

import os


class PSPContext:
    """PSPContext is an abstract base class for Context classes.

    Holds all the common stuff that various parts of the compilation
    will need access to. The items in this class will be used by both
    the compiler and the class generator.
    """

    def getClassPath(self):
        raise NotImplementedError

    def getReader(self):
        raise NotImplementedError

    def getWriter(self):
        raise NotImplementedError

    def getOutputDirectory(self):
        """Provide directory to dump PSP source file to."""
        raise NotImplementedError

    def getServletClassName(self):
        """Return the class name of the servlet being generated."""
        raise NotImplementedError

    def getFullClassName(self):
        """Return the class name including package prefixes.

        Won't use this for now.
        """
        raise NotImplementedError

    def getPythonFileName(self):
        """Return the filename that we are generating to."""
        raise NotImplementedError

    def getPythonFileEncoding(self):
        """Return the encoding of the file that we are generating."""
        raise NotImplementedError

    def setPSPReader(self, reader):
        """Set the PSPReader for this context."""
        raise NotImplementedError

    def setServletWriter(self, writer):
        """Set the PSPWriter instance for this context."""
        raise NotImplementedError

    def setPythonFileName(self, name):
        """Set the name of the .py file to generate."""
        raise NotImplementedError

    def setPythonFileEncoding(self, encoding):
        """Set the encoding of the .py file to generate."""
        raise NotImplementedError


class PSPCLContext(PSPContext):
    """A context for command line compilation.

    Currently used for both command line and PSPServletEngine compilation.
    This class provides all the information necessary during the parsing
    and page generation steps of the PSP compilation process.
    """

    def __init__(self, pspfile):
        PSPContext.__init__(self)
        self._baseUri, self._pspFile = os.path.split(pspfile)
        self._fullPath = pspfile
        self._pyFileEncoding = None

    def getClassPath(self):
        raise NotImplementedError

    def getReader(self):
        """Return the PSPReader object assigned to this context."""
        return self._pspReader

    def getServletWriter(self):
        """Return the ServletWriter object assigned to this context."""
        return self._servletWriter

    def getOutputDirectory(self):
        """Provide directory to dump PSP source file to.

        I am probably doing this in reverse order at the moment.
        I should start with this and get the Python filename from it.
        """
        return os.path.split(self._pyFileName)[0]

    def getServletClassName(self):
        """Return the class name of the servlet being generated."""
        return self._className

    def getFullClassName(self):
        """Return the class name including package prefixes.

        Won't use this for now.
        """
        raise NotImplementedError

    def getPythonFileName(self):
        """Return the filename that we are generating to."""
        return self._pyFileName

    def getPythonFileEncoding(self):
        """Return the encoding of the file that we are generating."""
        return self._pyFileEncoding

    def getPspFileName(self):
        """Return the name of the PSP file from which we are generating."""
        return self._pspFile

    def getFullPspFileName(self):
        """Return the name of the PSP file including its file path."""
        return self._fullPath

    def setPSPReader(self, reader):
        """Set the PSPReader for this context."""
        self._pspReader = reader

    def setServletWriter(self, writer):
        """Set the ServletWriter instance for this context."""
        self._servletWriter = writer

    def setPythonFileName(self, name):
        """Set the name of the .py file to generate."""
        self._pyFileName = name

    def setPythonFileEncoding(self, encoding):
        """Set the encoding of the .py file to generate."""
        self._pyFileEncoding = encoding

    def setClassName(self, name):
        """Set the class name to create."""
        self._className = name

    def resolveRelativeURI(self, uri):
        """This is used mainly for including files.

        It simply returns the location relative to the base context
        directory, ie Examples/. If the filename has a leading /,
        it is assumed to be an absolute path.
        """
        if os.path.isabs(uri):
            return uri
        return os.path.join(self._baseUri, uri)

    def getBaseUri(self):
        """Return the base URI for the servlet."""
        return self._baseUri



"""PickleCache.py

PickleCache provides tools for keeping fast-loading cached versions of
files so that subsequent loads are faster. This is similar to how Python
silently caches .pyc files next to .py files.

The typical scenario is that you have a type of text file that gets
"translated" to Pythonic data (dictionaries, tuples, instances, ints,
etc.). By caching the Python data on disk in pickle format, you can
avoid the expensive translation on subsequent reads of the file.

Two real life cases are MiscUtils.DataTable, which loads and represents
comma-separated files, and the separate MiddleKit plug-in which has an
object model file. So for examples on using this module, load up the
following files and search for "Pickle"::

    MiscUtils/DataTable.py
    MiddleKit/Core/Model.py

The cached file is named the same as the original file with
'.pickle.cache' suffixed. The utility of '.pickle' is to denote the file
format and the utility of '.cache' is to provide ``*.cache`` as a simple
pattern that can be removed, ignored by backup scripts, etc.

The treatment of the cached file is silent and friendly just like
Python's approach to .pyc files. If it cannot be read or written for
various reasons (cache is out of date, permissions are bad, wrong python
version, etc.), then it will be silently ignored.

GRANULARITY

In constructing the test suite, I discovered that if the source file is
newly written less than 1 second after the cached file, then the fact
that the source file is newer will not be detected and the cache will
still be used. I believe this is a limitation of the granularity of
os.path.getmtime(). If anyone knows of a more granular solution, please
let me know.

This would only be a problem in programmatic situations where the source
file was rapidly being written and read. I think that's fairly rare.

SEE ALSO
    https://docs.python.org/3/library/pickle.html
"""

import os
import sys
from time import sleep
from pprint import pprint
try:
    from pickle import load, dump, HIGHEST_PROTOCOL as maxPickleProtocol
except ImportError:
    from pickle import load, dump, HIGHEST_PROTOCOL as maxPickleProtocol

verbose = False

# force version_info into a simple tuple
versionInfo = tuple(sys.version_info)


class PickleCache:
    """Abstract base class for PickleCacheReader and PickleCacheWriter."""

    _verbose = verbose

    def picklePath(self, filename):
        return filename + '.pickle.cache'


class PickleCacheReader(PickleCache):

    def read(self, filename,
             pickleProtocol=None, source=None, verbose=None):
        """Read data from pickle cache.

        Returns the data from the pickle cache version of the filename,
        if it can read. Otherwise returns None, which also indicates
        that writePickleCache() should be subsequently called after
        the original file is read.
        """
        if pickleProtocol is None or pickleProtocol < 0:
            pickleProtocol = maxPickleProtocol
        if verbose is None:
            v = self._verbose
        else:
            v = verbose
        if v:
            print('>> PickleCacheReader.read() - verbose is on')
        if not filename:
            raise ValueError('Missing filename')

        if not os.path.exists(filename):
            if v:
                print(f'Cannot find {filename!r}.')
            open(filename)  # to get a properly constructed IOError

        shouldDeletePickle = False
        data = None

        picklePath = self.picklePath(filename)
        if os.path.exists(picklePath):
            if os.path.getmtime(picklePath) < os.path.getmtime(filename):
                if v:
                    print('Cache is out of date.')
                shouldDeletePickle = True
            else:
                try:
                    if v:
                        print(f'About to open for read {picklePath!r}.')
                    file = open(picklePath, 'rb')
                except IOError as e:
                    if v:
                        print('Cannot open cache file:'
                              f' {e.__class__.__name__}: {e}.')
                else:
                    try:
                        if v:
                            print('about to load')
                        d = load(file)
                    except EOFError:
                        if v:
                            print('EOFError - not loading')
                        shouldDeletePickle = True
                    except Exception as exc:
                        print(f'WARNING: {self.__class__.__name__}:'
                              f' {exc.__class__.__name__}: {exc}')
                        shouldDeletePickle = True
                    else:
                        file.close()
                        if v:
                            print('Finished reading.')
                        if not isinstance(d, dict):
                            raise TypeError(f'Expected dict, but got: {d!r}')
                        for key in ('source', 'data',
                                    'pickle protocol', 'python version'):
                            if key not in d:
                                raise ValueError(f'{key} not found')
                        if source and d['source'] != source:
                            if v:
                                print(f'Not from required source ({source}):'
                                      f" {d['source']}.")
                            shouldDeletePickle = True
                        elif d['pickle protocol'] != pickleProtocol:
                            if v:
                                print('Pickle protocol'
                                      f" ({d['pickle protocol']})"
                                      ' does not match expected'
                                      f' ({pickleProtocol}).')
                            shouldDeletePickle = True
                        elif d['python version'] != versionInfo:
                            if v:
                                print('Python version'
                                      " {d['python version']}"
                                      ' does not match current'
                                      ' {versionInfo}')
                            shouldDeletePickle = True
                        else:
                            if v:
                                print('All tests pass, accepting data.')
                                if v > 1:
                                    print('Display full dict:')
                                    pprint(d)
                            data = d['data']

        # Delete the pickle file if suggested by previous conditions
        if shouldDeletePickle:
            try:
                if v:
                    print('Attempting to remove pickle cache file.')
                os.remove(picklePath)
            except OSError as exc:
                if v:
                    print(f'Failed to remove: {exc.__class__.__name__}: {exc}')

        if v:
            print('Done reading data.')
            print()

        return data


class PickleCacheWriter(PickleCache):

    _writeSleepInterval = 0.1

    def write(self, data, filename,
              pickleProtocol=None, source=None, verbose=None):
        if pickleProtocol is None or pickleProtocol < 0:
            pickleProtocol = maxPickleProtocol
        if verbose is None:
            v = self._verbose
        else:
            v = verbose
        if v:
            print('>> PickleCacheWriter.write() - verbose is on.')
        if not filename:
            raise ValueError('Missing filename')
        sourceTimestamp = os.path.getmtime(filename)

        picklePath = self.picklePath(filename)
        d = {
            'source': source,
            'python version': versionInfo,
            'pickle protocol': pickleProtocol,
            'data': data,
        }
        if v > 1:
            print('Display full dict:')
            pprint(d)
        try:
            if v:
                print('About to open for write %r.' % picklePath)
            pickleFile = open(picklePath, 'wb')
        except IOError as e:
            if v:
                print(f'error. not writing. {e.__class__.__name__}: {e}')
        else:
            while True:
                dump(d, pickleFile, pickleProtocol)
                pickleFile.close()
                # Make sure the cache has a newer timestamp, otherwise the
                # cache will just get ignored and rewritten next time.
                if os.path.getmtime(picklePath) == sourceTimestamp:
                    if v:
                        print('Timestamps are identical, sleeping'
                              f' {self._writeSleepInterval:%0.2f} seconds.')
                    sleep(self._writeSleepInterval)
                    pickleFile = open(picklePath, 'wb')
                else:
                    break

        if v:
            print('Done writing data.')
            print()


# Define module level convenience functions:
_reader = PickleCacheReader()
readPickleCache = _reader.read
_writer = PickleCacheWriter()
writePickleCache = _writer.write



"""HTMLForException.py

Create HTML for exceptions.
"""

import os
import re
import sys
import traceback
import urllib.request
import urllib.parse
import urllib.error

from .Funcs import htmlEncode

HTMLForExceptionOptions = {
    'table': 'background-color:#f0f0f0',
    'default': 'color:#000',
    'row.location': 'color:#009',
    'row.code': 'color:#900',
    'editlink': None,
}

fileRE = re.compile(r'File "([^"]*)", line ([0-9]+), in ([^ ]*)')


def htmlForLines(lines, options=None):
    """Create HTML for exceptions and tracebacks from a list of strings."""

    # Set up the options:
    if options:
        opt = HTMLForExceptionOptions.copy()
        opt.update(options)
    else:
        opt = HTMLForExceptionOptions

    # Create the HTML:
    res = ['<table style="width:100%;{}">\n'.format(opt['table']),
           '<tr><td><pre style="{}">\n'.format(opt['default'])]
    for line in lines:
        match = fileRE.search(line)
        if match:
            parts = list(map(htmlEncode, line.split('\n', 2)))
            parts[0] = '<span style="{}">{}</span>'.format(
                opt['row.location'], parts[0])
            if opt['editlink']:
                parts[0] = (
                    '{} <a href="{}?filename={}&amp;line={}">'
                    '[edit]</a>'.format(
                        parts[0], opt['editlink'], urllib.parse.quote(
                            os.path.abspath(match.group(1))), match.group(2)))
            parts[1] = '<span style="{}">{}</span>'.format(
                opt['row.code'], parts[1])
            line = '\n'.join(parts)
            res.append(line)
        else:
            res.append(htmlEncode(line))
    if lines:
        if res[-1][-1] == '\n':
            res[-1] = res[-1].rstrip()
    res.extend(['</pre></td></tr>\n', '</table>\n'])
    return ''.join(res)


def htmlForStackTrace(frame=None, options=None):
    """Get HTML for displaying a stack trace.

    Returns an HTML string that presents useful information to the developer
    about the stack. The first argument is a stack frame such as returned by
    sys._getframe() which is in fact invoked if a stack frame isn't provided.
    """

    # Get the stack frame if needed:
    if frame is None:
        frame = sys._getframe()

    # Return formatted stack traceback
    return htmlForLines(traceback.format_stack(frame), options)


def htmlForException(excInfo=None, options=None):
    """Get HTML for displaying an exception.

    Returns an HTML string that presents useful information to the developer
    about the exception. The first argument is a tuple such as returned by
    sys.exc_info() which is in fact invoked if the tuple isn't provided.
    """

    # Get the excInfo if needed:
    if excInfo is None:
        excInfo = sys.exc_info()

    # Return formatted exception traceback
    return htmlForLines(traceback.format_exception(*excInfo), options)


# old (deprecated) aliases
HTMLForLines, HTMLForStackTrace, HTMLForException = (
    htmlForLines, htmlForStackTrace, htmlForException)



import unittest

from MiscUtils.CSVParser import CSVParser, ParseError


class CSVParserTests(unittest.TestCase):
    """CSV parser tests.

    TO DO

      * Test the different options for parser, see CSVParser.__init__().
    """

    def setUp(self):
        self.parse = CSVParser().parse

    def testNegatives(self):
        inputs = [
            '""a',
            '"a"b',
            'a\n,b'
        ]
        for inp in inputs:
            try:
                results = self.parse(inp)
            except ParseError:
                pass
            else:
                print()
                print('results:', repr(results))
                raise Exception(f'Did not get an exception for: {inp!r}')

    def testPositives(self):
        tests = [
            # basics
            ('', []),
            (',', ['', '']),
            (',,', ['', '', '']),
            ('a', ['a']),
            ('a,b', ['a', 'b']),
            ('a,b,c,d,e,f', 'a b c d e f'.split()),

            # surrounding whitespace
            (' a', ['a']),
            ('a ', ['a']),
            (' a ', ['a']),
            ('a, b', ['a', 'b']),
            ('  a  ,  b  ', ['a', 'b']),

            # commas in fields
            ('","', [',']),
            ('",",","', [',', ',']),
            ('"a  ,  b",b', ['a  ,  b', 'b']),

            # quotes in fields
            ('""""', ['"']),
            ('""""""', ['""']),
            ('"""a""",b,"""c"""', ['"a"', 'b', '"c"']),

            # single line combos
            (' "a", "b"', ['a', 'b']),
            ('  """"', ['"']),
            ('""""  ', ['"']),
            ('  """"  ', ['"']),
            (' """a""",  """b"""', ['"a"', '"b"']),
            ('  """",  ",",   ""","""', ['"', ',', '","']),

            # comments
            ('#a,b', []),

            # multiple line records
            ('"a\nb"', ['a\nb']),
            ('a,"b\nc"', ['a', 'b\nc']),
            ('a,"b\nc\n\n\n"', ['a', 'b\nc']),

            # MiddleKit enums
            ('a,Enums="b"', ['a', 'Enums="b"']),
            ("a,Enums='b'", ['a', "Enums='b'"]),
            ('a,"Enums=""b, c"""', ['a', 'Enums="b, c"']),
            ('''a,"Enums='b, c'"''', ['a', "Enums='b, c'"]),
        ]
        for inp, out in tests:
            if '\n' not in inp:
                # single line
                res = self.parse(inp)
                self.assertEqual(
                    res, out,
                    f'\ninput={inp!r}\nresult={res!r}\noutput={out!r}')
                res = self.parse(inp+'\n')
                self.assertEqual(
                    res, out,
                    f'\ninput={inp!r}\nresult={res!r}\noutput={out!r}')
            else:
                # multiple lines
                gotFields = False
                for line in inp.splitlines():
                    self.assertFalse(gotFields)
                    res = self.parse(line)
                    if res is not None:
                        gotFields = True
                self.assertTrue(gotFields)
                self.assertEqual(
                    res, out,
                    f'\ninput={inp!r}\nresult={res!r}\noutput={out!r}')



import datetime
import time

from SidebarPage import SidebarPage

cookieValues = [
    ('onclose', 'ONCLOSE'),
    ('expireNow', 'NOW'),
    ('expireNever', 'NEVER'),
    ('oneMinute', '+1m'),
    ('oneWeek', '+1w'),
    ('oneHourAndHalf', '+ 1h 30m'),
    ('timeIntTenSec', time.time() + 10),
    ('tupleOneYear', (time.localtime()[0] + 1,) + time.localtime()[1:]),
    ('dt1day', datetime.datetime.now() + datetime.timedelta(1)),
    ('dt2min', datetime.timedelta(minutes=2))
]

cookieIndex = 1


class SetCookie(SidebarPage):

    def cornerTitle(self):
        return 'Testing'

    def writeContent(self):
        global cookieIndex
        res = self.response()
        req = self.request()
        enc = self.htmlEncode
        self.writeln('<h4>The time right now is:</h4>')
        t = time.strftime('%a, %d-%b-%Y %H:%M:%S GMT', time.gmtime())
        self.writeln(f'<p>{t}</p>')
        self.writeln('<h2>Cookies received:</h2>\n')
        self.writeln('<ul>')
        for name, value in req.cookies().items():
            self.writeln(f'<li>{name!r} = {enc(value)}</li>')
        self.writeln('</ul>')
        for name, expire in cookieValues:
            res.setCookie(name, f'Value #{cookieIndex}', expires=expire)
            cookieIndex += 1
        self.writeln('<h2>Cookies being sent:</h2>\n')
        self.writeln('<dl>')
        for name, cookie in res.cookies().items():
            self.writeln(f'<dt>{name!r} sends:</dt>')
            self.writeln(f'<dd>{enc(cookie.headerValue())}</dd>')
        self.writeln('</dl>')



"""More comprehensive traceback formatting for Python scripts.

Original version know as cgitb written By Ka-Ping Yee <ping@lfw.org>
Modified for Webware by Ian Bicking <ianb@colorstudy.com>
"""

import inspect
import keyword
import linecache
import os
import pydoc
import sys
import tokenize
from types import MethodType

pyhtml = pydoc.html
escape = pyhtml.escape

DefaultOptions = {
    'table': 'background-color:#f0f0f0',
    'default': 'color:#000',
    'row.location': 'color:#009',
    'row.code': 'color:#900',
    'header': 'color:#fff;background-color:#999',
    'subheader': 'color:#000;background-color:#f0f0f0;font-size:10pt',
    'code.accent': 'background-color:#ffc',
    'code.unaccent': 'color:#999;font-size:10pt',
}


def breaker():
    return (
        '<body style="background-color:#f0f0ff">' +
        '<span style="color:#F0F0FF;font-size:small"> > </span> ' +
        '</table>' * 5)


def html(context=5, options=None):
    if options:
        opt = DefaultOptions.copy()
        opt.update(options)
    else:
        opt = DefaultOptions

    etype, evalue = sys.exc_info()[:2]
    if not isinstance(etype, str):
        etype = etype.__name__
    inspect_trace = inspect.trace(context)

    javascript = """
    <script>
    function tag(s) { return '<'+s+'>'; }
    function popup_repr(title, value) {
        w = window.open('', '_blank',
            'directories=no,height=240,width=480,location=no,menubar=yes,'
            +'resizable=yes,scrollbars=yes,status=no,toolbar=no');
        if (!w) return true;
        w.document.open();
        w.document.write(tag('html')+tag('head')
            +tag('title')+title+tag('/title')+tag('/head')
            +tag('body style="background-color:#fff"')
            +tag('h3')+title+':'+tag('/h3')
            +tag('p')+tag('code')+value+tag('/code')+tag('/p')+tag('form')+
            tag('input type="button" onClick="window.close()" value="Close"')
            +tag('/form')+tag('/body')+tag('/html'));
        w.document.close();
        return false;
    }
    </script>
    """

    traceback_summary = []

    for frame, filename, lineno, func, lines, index in reversed(inspect_trace):
        if filename:
            filename = os.path.abspath(filename)
        else:
            filename = 'not found'
        traceback_summary.append(
            '<a href="#{}:{}" style="{}">{}</a>:'
            '<code style="font-family:Courier,sans-serif">{}</code>'.format(
                filename.replace('/', '-').replace('\\', '-'), lineno,
                opt['header'], os.path.splitext(os.path.basename(filename))[0],
                '{:5d}'.format(lineno).replace(' ', '&nbsp;')))

    head = (
        '<table style="width:100%;{}">'
        '<tr><td style="text-align:left;vertical-align:top">'
        '<strong style="font-size:x-large">{}</strong>: {}</td><td rowspan'
        '="2" style="text-align:right;vertical-align:top">{}</td></tr>'
        '<tr><td style="vertical-align:top;background-color:#fff">\n'
        '<p style="{}">A problem occurred while running a Python script.</p>'
        '<p style="{}">Here is the sequence of function calls leading up to'
        ' the error, with the most recent (innermost) call first.</p>\n'
        '</td></tr></table>\n'.format(
            opt['header'], etype, escape(str(evalue)),
            '<br>\n'.join(traceback_summary), opt['default'], opt['default']))

    indent = '<code><small>{}</small>&nbsp;</code>'.format('&nbsp;' * 5)
    traceback = []
    for frame, filename, lineno, func, lines, index in reversed(inspect_trace):
        if filename:
            filename = os.path.abspath(filename)
        else:
            filename = 'not found'
        try:
            file_list = filename.split('/')
            display_file = '/'.join(
                file_list[file_list.index('Webware') + 1:])
        except ValueError:
            display_file = filename
        if display_file[-3:] == '.py':
            display_file = display_file[:-3]
        link = '<a id="{}:{}"></a><a href="file:{}">{}</a>'.format(
            filename.replace('/', '-').replace('\\', '-'),
            lineno, filename.replace('\\', '/'), escape(display_file))
        args, varargs, varkw, locals = inspect.getargvalues(frame)
        if func == '?':
            call = ''
        else:
            call = 'in <strong>{}</strong>'.format(
                func + inspect.formatargvalues(
                    args, varargs, varkw, locals,
                    formatvalue=lambda value: '=' + html_repr(value)))

        names = []
        dotted = [0, []]

        def tokeneater():
            if type_ == tokenize.OP and token == '.':
                dotted[0] = 1
            if type_ == tokenize.NAME and token not in keyword.kwlist:
                if dotted[0]:
                    dotted[0] = 0
                    dotted[1].append(token)
                    if token not in names:
                        names.append(dotted[1][:])
                elif token not in names:
                    if token != 'self':
                        names.append(token)
                    dotted[1] = [token]
            if type_ == tokenize.NEWLINE:
                raise IndexError

        def linereader():
            nonlocal lineno
            line = linecache.getline(filename, lineno)
            line = line.encode('utf-8')
            lineno += 1
            return line

        for type_, token, _start, _end, _line in tokenize.tokenize(linereader):
            try:
                tokeneater()
            except IndexError:
                break

        lvals = []
        for name in names:
            if isinstance(name, list):
                if name[0] in locals or name[0] in frame.f_globals:
                    name_list, name = name, name[0]
                    if name_list[0] in locals:
                        value = locals[name_list[0]]
                    else:
                        value = frame.f_globals[name_list[0]]
                        name = f'<em>global</em> {name}'
                    for subname in name_list[1:]:
                        if hasattr(value, subname):
                            value = getattr(value, subname)
                            name += '.' + subname
                        else:
                            name += f'.(unknown: {name})'
                            break
                    name = f'<strong>{name}</strong>'
                    if isinstance(value, MethodType):
                        value = None
                    else:
                        value = html_repr(value)
            elif name in frame.f_code.co_varnames:
                if name in locals:
                    value = html_repr(locals[name])
                else:
                    value = '<em>undefined</em>'
                name = f'<strong>{name}</strong>'
            else:
                if name in frame.f_globals:
                    value = html_repr(frame.f_globals[name])
                else:
                    value = '<em>undefined</em>'
                name = f'<em>global</em> <strong>{name}</strong>'
            if value is not None:
                lvals.append(f'{name}&nbsp;= {value}')
        if lvals:
            lvals = ', '.join(lvals)
            lvals = indent + '<span style="{}">{}</span><br>\n'.format(
                opt['code.unaccent'], lvals)
        else:
            lvals = ''

        level = (
            '<br><table style="width:100%;{}">'
            '<tr><td>{} {}</td></tr></table>'.format(
                opt['subheader'], link, call))
        excerpt = []
        try:
            i = lineno - index
        except TypeError:
            i = lineno
        lines = lines or ['file not found']
        for line in lines:
            number = '&nbsp;' * (5-len(str(i))) + str(i)
            number = '<span style="{}">{}</span>'.format(
                opt['code.unaccent'], number)
            line = '<code>{}&nbsp;{}</code>'.format(
                number, pyhtml.preformat(line))
            if i == lineno:
                line = (
                    '<table style="width:100%;{}">'
                    '<tr><td>{}</td></tr></table>'.format(
                        opt['code.accent'], line))
            excerpt.append('\n' + line)
            if i == lineno:
                excerpt.append(lvals)
            i += 1
        traceback.append(level + '\n'.join(excerpt))

    exception = '<p><strong>{}</strong>: {}\n'.format(
        etype, escape(str(evalue)))
    attribs = []
    if evalue is not None:
        for name in dir(evalue):
            if name.startswith('__'):
                continue
            value = html_repr(getattr(evalue, name))
            attribs.append(f'<br>{indent}{name}&nbsp;= {value}\n')
    return (javascript + head + ''.join(traceback) +
            exception + ''.join(attribs) + '</p>\n')


def handler():
    print(breaker())
    print(html())


def html_repr(value):
    html_repr_instance = pyhtml._repr_instance
    enc_value = pyhtml.repr(value)
    if len(enc_value) > html_repr_instance.maxstring:
        plain_value = escape(repr(value))
        return (
            '{} <a href="#" onClick="return popup_repr('
            "'Full representation','{}')"
            '" title="Full representation">(complete)</a>'.format(
                enc_value, escape(plain_value).replace(
                    "'", "\\'").replace('"', '&quot;')))
    else:
        return enc_value



import sys

from SidebarPage import SidebarPage

modName = __name__  # should be Testing.ServletImport


class ServletImport(SidebarPage):
    """Test of servlet import details."""

    def cornerTitle(self):
        return 'Testing'

    def writeContent(self):
        wr = self.writeln
        wr('<h2>Webware Servlet Import Test</h2>')
        wr(f'<h3>{self.__doc__}</h3>')
        modNameFromClass = ServletImport.__module__
        modNameConsistent = modName == modNameFromClass
        servletInSysModules = modName in sys.modules
        app = self.transaction().application()
        files = app._imp.fileList(update=False)
        servletFileWatched = __file__ in files
        wr(
            f"<p>modName = <code>{modName}</code></p>"
            f"<p>modNameFromClass = <code>{modNameFromClass}</code></p>"
            f"<p>modNameConsistent = <code>{modNameConsistent}</code></p>"
            f"<p>servletInSysModules = <code>{servletInSysModules}</code></p>"
            f"<p>servletFileWatched = <code>{servletFileWatched}</code></p>"
        )
        ok = modNameConsistent and servletInSysModules and servletInSysModules
        wr('<p style="color:{}"><b>Servlet import test {}.</b></p>'.format(
            'green' if ok else 'red', 'passed' if ok else 'failed'))



import unittest

from MiscUtils.DateInterval import timeEncode, timeDecode


class TestDateInterval(unittest.TestCase):

    def testTimeEncode(self):
        self.assertEqual(timeEncode(1), '1s')
        self.assertEqual(timeEncode(60), '1m')
        self.assertEqual(timeEncode(176403), '2d1h3s')
        self.assertEqual(timeEncode(349380), '4d1h3m')
        self.assertEqual(timeEncode(38898367), '1y2b3w4d5h6m7s')

    def testTimeDecode(self):
        self.assertEqual(timeDecode('1s'), 1)
        self.assertEqual(timeDecode('1h2d3s'), 176403)
        self.assertEqual(timeDecode('2d1h3s'), 176403)
        self.assertEqual(timeDecode('1h4d3m'), 349380)
        self.assertEqual(timeDecode('3m4d1h'), 349380)
        self.assertEqual(timeDecode('1y2b3w4d5h6m7s'), 38898367)
        self.assertEqual(timeDecode('0y1b2w3d4h5m6s'), 4075506)
        self.assertEqual(timeDecode('6s5m4h3d2w1b0y'), 4075506)
        self.assertEqual(timeDecode('(3s-2d-1h)'), 176403)
        try:
            timeDecode('1h5n')
        except ValueError as e:
            self.assertEqual(str(e), 'Invalid unit of time: n')



"""MiscUtils.Funcs

This module holds functions that don't fit in anywhere else.

You can safely import * from MiscUtils.Funcs if you like.
"""

import os
import random
import datetime
import time
import textwrap

from hashlib import sha256
from struct import calcsize, pack

__all__ = [
    'commas', 'charWrap', 'wordWrap', 'excstr', 'hostName', 'localIP',
    'positiveId', 'safeDescription', 'asclocaltime', 'timestamp',
    'localTimeDelta', 'uniqueId', 'valueForString']


def commas(number):
    """Insert commas in a number.

    Return the given number as a string with commas to separate
    the thousands positions.

    The number can be a float, int, long or string. Returns None for None.
    """
    if number is None:
        return None
    if not number:
        return str(number)
    number = list(str(number))
    if '.' in number:
        i = number.index('.')
    else:
        i = len(number)
    while True:
        i -= 3
        if i <= 0 or number[i - 1] == '-':
            break
        number.insert(i, ',')
    return ''.join(number)


def charWrap(s, width, hanging=0):
    """Word wrap a string.

    Return a new version of the string word wrapped with the given width
    and hanging indent. The font is assumed to be monospaced.

    This can be useful for including text between ``<pre>...</pre>`` tags,
    since ``<pre>`` will not word wrap, and for lengthy lines, will increase
    the width of a web page.

    It can also be used to help delineate the entries in log-style
    output by passing hanging=4.
    """
    if not s:
        return s
    if hanging < 0 or width < 1 or hanging >= width:
        raise ValueError("Invalid width or indentation")
    hanging = ' ' * hanging
    lines = s.splitlines()
    i = 0
    while i < len(lines):
        s = lines[i]
        while len(s) > width:
            lines[i], s = s[:width].rstrip(), hanging + s[width:].lstrip()
            i += 1
            lines.insert(i, s)
        i += 1
    return '\n'.join(lines)


def wordWrap(s, width=78):
    """Return a version of the string word wrapped to the given width."""
    return textwrap.fill(s, width)


def excstr(e):
    """Return a string for the exception.

    The string will be in the format that Python normally outputs
    in interactive shells and such::

        <ExceptionName>: <message>
        AttributeError: 'object' object has no attribute 'bar'

    Neither str(e) nor repr(e) do that.
    """
    if e is None:
        return None
    return f'{e.__class__.__name__}: {e}'


def hostName():
    """Return the host name.

    The name is taken first from the os environment and failing that,
    from the 'hostname' executable. May return None if neither attempt
    succeeded. The environment keys checked are HOST and HOSTNAME,
    both upper and lower case.
    """
    get = os.environ.get
    for name in ('host', 'hostname', 'computername'):
        name = get(name.upper()) or get(name)
        if name:
            break
    if not name:
        with os.popen('hostname') as f:
            name = f.read().strip()
    if name:
        name = name.lower()
    return name or None


_localIP = None


def localIP(remote=('www.yahoo.com', 80), useCache=True):
    """Get the "public" address of the local machine.

    This is the address which is connected to the general Internet.

    This function connects to a remote HTTP server the first time it is
    invoked (or every time it is invoked with useCache=0). If that is
    not acceptable, pass remote=None, but be warned that the result is
    less likely to be externally visible.

    Getting your local ip is actually quite complex. If this function
    is not serving your needs then you probably need to think deeply
    about what you really want and how your network is really set up.
    Search comp.lang.python for "local ip" for more information.
    """
    global _localIP
    if useCache and _localIP:
        return _localIP
    import socket
    if remote:
        # code from Donn Cave on comp.lang.python
        #
        # Why not use this? socket.gethostbyname(socket.gethostname())
        # On some machines, it returns '127.0.0.1' - not what we had in mind.
        #
        # Why not use this? socket.gethostbyname_ex(socket.gethostname())[2]
        # Because some machines have more than one IP (think "VPN", etc.) and
        # there is no easy way to tell which one is the externally visible IP.
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.connect(remote)
                address, _port = s.getsockname()
        except socket.error:
            pass  # oh, well, we'll use the local method
        else:
            if address and not address.startswith('127.'):
                if useCache:
                    _localIP = address
                return address
    try:
        hostname = socket.gethostname()
        addresses = socket.gethostbyname_ex(hostname)[2]
    except socket.error:
        addresses = socket.gethostbyname_ex('localhost')[2]
    for address in addresses:
        if address and not address.startswith('127.'):
            if useCache:
                _localIP = address
            return address
    if useCache:
        _localIP = addresses[0]
    return _localIP


# Addresses can "look negative" on some boxes, some of the time.
# If you feed a "negative address" to an %x format, modern Python
# versions will display it as signed. So when you want to produce
# an address, use positiveId() to obtain it.
# _address_mask is 2**(number_of_bits_in_a_native_pointer).
# Adding this to a negative address gives a positive int with the same
# hex representation as the significant bits in the original.
# This idea and code were taken from ZODB
# (https://github.com/zopefoundation/ZODB).

_address_mask = 256 ** calcsize('P')


def positiveId(obj):
    """Return id(obj) as a non-negative integer."""
    result = id(obj)
    if result < 0:
        result += _address_mask
        if not result > 0:
            raise ValueError('Cannot generate a non-negative integer')
    return result


def _descExc(reprOfWhat, err):
    """Return a description of an exception.

    This is a private function for use by safeDescription().
    """
    try:
        return (f'(exception from repr({reprOfWhat}):'
                f' {err.__class__.__name__}: {err})')
    except Exception:
        return f'(exception from repr({reprOfWhat}))'


def safeDescription(obj, what='what'):
    """Return the repr() of obj and its class (or type) for help in debugging.

    A major benefit here is that exceptions from repr() are consumed.
    This is important in places like "assert" where you don't want
    to lose the assertion exception in your attempt to get more information.

    Example use::

        assert isinstance(foo, Foo), safeDescription(foo)
        print("foo:", safeDescription(foo))  # won't raise exceptions

        # better output format:
        assert isinstance(foo, Foo), safeDescription(foo, 'foo')
        print(safeDescription(foo, 'foo'))
    """
    try:
        xRepr = repr(obj)
    except Exception as e:
        xRepr = _descExc('obj', e)
    if hasattr(obj, '__class__'):
        try:
            cRepr = repr(obj.__class__)
        except Exception as e:
            cRepr = _descExc('obj.__class__', e)
        return f'{what}={xRepr} class={cRepr}'
    try:
        cRepr = repr(type(obj))
    except Exception as e:
        cRepr = _descExc('type(obj)', e)
    return f'{what}={xRepr} type={cRepr}'


def asclocaltime(t=None):
    """Return a readable string of the current, local time.

    Useful for time stamps in log files.
    """
    return time.asctime(time.localtime(t))


def timestamp(t=None):
    """Return a dictionary whose keys give different versions of the timestamp.

    The dictionary will contain the following timestamp versions::

        'tuple': (year, month, day, hour, min, sec)
        'pretty': 'YYYY-MM-DD HH:MM:SS'
        'condensed': 'YYYYMMDDHHMMSS'
        'dashed': 'YYYY-MM-DD-HH-MM-SS'

    The focus is on the year, month, day, hour and second, with no additional
    information such as timezone or day of year. This form of timestamp is
    often ideal for print statements, logs and filenames. If the current number
    of seconds is not passed, then the current time is taken. The 'pretty'
    format is ideal for print statements, while the 'condensed' and 'dashed'
    formats are generally more appropriate for filenames.
    """
    t = time.localtime(t)[:6]
    return dict(
        tuple=t,
        pretty='{:4d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}'.format(*t),
        condensed='{:4d}{:02d}{:02d}{:02d}{:02d}{:02d}'.format(*t),
        dashed='{:4d}-{:02d}-{:02d}-{:02d}-{:02d}-{:02d}'.format(*t))


def localTimeDelta(t=None):
    """Return timedelta of local zone from GMT."""
    if t is None:
        t = time.time()
    dt = datetime.datetime
    return dt.fromtimestamp(t) - dt.utcfromtimestamp(t)


def uniqueId(forObject=None):
    """Generate an opaque identifier string made of 32 hex digits.

    The string is practically guaranteed to be unique for each call.

    If a randomness source is not found in the operating system, this function
    will use SHA-256 hashing with a combination of pseudo-random numbers and
    time values to generate additional randomness. In this case, if an object
    is passed, then its id() will be incorporated into the generation as well.
    """
    try:  # prefer os.urandom(), if available
        return os.urandom(16).hex()
    except (AttributeError, NotImplementedError):
        r = pack('3f', time.time(), random.random(), os.times()[0])
        if forObject is not None:
            r += pack('q', id(forObject))
        return sha256(r).hexdigest()[:32]


def valueForString(s):
    """Return value for a string.

    For a given string, returns the most appropriate Pythonic value
    such as None, a long, an int, a list, etc. If none of those
    make sense, then returns the string as-is.

    "None", "True" and "False" are case-insensitive because there is
    already too much case sensitivity in computing, damn it!
    """
    if not s:
        return s
    try:
        return int(s)
    except ValueError:
        pass
    try:
        return int(s)
    except ValueError:
        pass
    try:
        return float(s)
    except ValueError:
        pass
    t = s.lower()
    if t == 'none':
        return None
    if t == 'true':
        return True
    if t == 'false':
        return False
    if s[0] in '[({"\'':
        return eval(s)
    return s



import unittest

from MiscUtils.Error import Error


class DummyObject:

    def __repr__(self):
        return '<dummy>'


class TestError(unittest.TestCase):

    def testNone(self):
        err = Error(None, None)
        self.assertEqual(str(err), 'ERROR: None')
        self.assertEqual(repr(err),
                         'ERROR(object=None; message=None; data={})')
        self.assertIs(err.object(), None)
        self.assertIs(err.message(), None)

    def testObjMessage(self):
        obj = DummyObject()
        err = Error(obj, 'test')
        self.assertEqual(str(err), 'ERROR: test')
        self.assertEqual(
            repr(err),
            "ERROR(object=<dummy>; message='test'; data={})")
        self.assertIs(err.object(), obj)
        self.assertIs(err.message(), 'test')

    def testIsDict(self):
        err = Error(DummyObject(), 'test')
        self.assertIsInstance(err, dict)
        self.assertIsInstance(err, Error)

    def testValueDict(self):
        err = Error(DummyObject(), 'test', a=5, b='.')
        self.assertEqual(str(err), 'ERROR: test')
        self.assertEqual(
            repr(err),
            "ERROR(object=<dummy>; message='test'; data={'a': 5, 'b': '.'})")
        self.assertEqual(list(err), ['a', 'b'])
        self.assertIsInstance(err['a'], int)
        self.assertIsInstance(err['b'], str)

    def testVarArgs(self):
        err = Error(DummyObject(), 'test', {'a': 5}, b='.')
        self.assertEqual(str(err), 'ERROR: test')
        self.assertEqual(
            repr(err),
            "ERROR(object=<dummy>; message='test'; data={'a': 5, 'b': '.'})")
        self.assertEqual(list(err), ['a', 'b'])
        self.assertIsInstance(err['a'], int)
        self.assertIsInstance(err['b'], str)



"""BraceConverter.py

Contributed 2000-09-04 by Dave Wallace (dwallace@delanet.com)

Converts Brace-blocked Python into normal indented Python.
Brace-blocked Python is non-indentation aware and blocks are
delimited by ':{' and '}' pairs.

Thus::

    for x in range(10) :{
        if x % 2 :{ print(x) } else :{ print(z) }
    }

Becomes (roughly, barring some spurious whitespace)::

    for x in range(10):
        if x % 2:
            print(x)
        else:
            print(z)

This implementation is fed a line at a time via parseLine(),
outputs to a PSPServletWriter, and tracks the current quotation
and block levels internally.
"""

import re


class BraceConverter:

    _reCharSkip = re.compile("(^[^\"'{}:#]+)")
    _reColonBrace = re.compile(r":\s*{\s*([^\s].*)?$")

    def __init__(self):
        self.inQuote = False
        self.dictLevel = 0

    def parseLine(self, line, writer):
        """Parse a line.

        The only public method of this class, call with subsequent lines
        and an instance of PSPServletWriter.
        """
        self.line = line
        if self.inQuote and self.line:
            self.skipQuote(writer)
        self.line = self.line.lstrip()
        if not self.line:
            writer.printChars('\n')
            return
        writer.printIndent()
        while self.line:
            while self.inQuote and self.line:
                self.skipQuote(writer)
            match = self._reCharSkip.search(self.line)
            if match:
                writer.printChars(self.line[:match.end(1)])
                self.line = self.line[match.end(1):]
            else:
                c = self.line[0]
                if c == "'":
                    self.handleQuote("'", writer)
                    self.skipQuote(writer)
                elif c == '"':
                    self.handleQuote('"', writer)
                    self.skipQuote(writer)
                elif c == '{':
                    self.openBrace(writer)
                elif c == '}':
                    self.closeBrace(writer)
                elif c == ':':
                    self.openBlock(writer)
                elif c == '#':
                    writer.printChars(self.line)
                    self.line = ''
                else:
                    # should never get here
                    raise Exception()
        writer.printChars('\n')

    def openBlock(self, writer):
        """Open a new block."""
        match = self._reColonBrace.match(self.line)
        if match and not self.dictLevel:
            writer.printChars(':')
            writer.pushIndent()
            if match.group(1):
                # text follows :{, if it's a comment, leave it on same line,
                # else start a new line and leave the text for processing
                if match.group(1)[0] == '#':
                    writer.printChars(' ' + match.group(1))
                    self.line = ''
                else:
                    writer.printChars('\n')
                    writer.printIndent()
                    self.line = match.group(1)
            else:
                self.line = ''
        else:
            writer.printChars(':')
            self.line = self.line[1:]

    def openBrace(self, writer):
        """Open brace encountered."""
        writer.printChars('{')
        self.line = self.line[1:]
        self.dictLevel += 1

    def closeBrace(self, writer):
        """Close brace encountered."""
        if self.dictLevel:
            writer.printChars('}')
            self.line = self.line[1:]
            self.dictLevel -= 1
        else:
            writer.popIndent()
            self.line = self.line[1:].lstrip()
            if self.line:
                writer.printChars('\n')
                writer.printIndent()

    def skipQuote(self, writer):
        """Skip to end of quote.

        Skip over all chars until the line is exhausted
        or the current non-escaped quote sequence is encountered.
        """
        pos = self.line.find(self.quoteChars)
        if pos < 0:
            writer.printChars(self.line)
            self.line = ''
        elif pos > 0 and self.line[pos - 1] == '\\':
            pos += 1
            writer.printChars(self.line[:pos])
            self.line = self.line[pos:]
            self.skipQuote(writer)
        else:
            pos += len(self.quoteChars)
            writer.printChars(self.line[:pos])
            self.line = self.line[pos:]
            self.inQuote = False

    def handleQuote(self, quote, writer):
        """Check and handle if current pos is a single or triple quote."""
        self.inQuote = True
        triple = quote * 3
        if self.line[0:3] == triple:
            self.quoteChars = triple
            writer.printChars(triple)
            self.line = self.line[3:]
        else:
            self.quoteChars = quote
            writer.printChars(quote)
            self.line = self.line[1:]



from SessionDynamicStore import SessionDynamicStore

from .TestSessionMemoryStore import TestSessionMemoryStore


class SessionDynamicStoreTest(TestSessionMemoryStore):

    _storeClass = SessionDynamicStore

    def testCleanStaleSessions(self):
        store = self._store
        memoryStore = store._memoryStore  # pylint: disable=no-member
        fileStore = store._fileStore  # pylint: disable=no-member
        self.assertEqual(len(memoryStore), 7)
        self.assertEqual(len(fileStore), 0)
        TestSessionMemoryStore.testCleanStaleSessions(self)
        self.assertEqual(len(memoryStore), 3)
        self.assertEqual(len(fileStore), 2)
        self.assertTrue('foo-0' in memoryStore and 'foo-2' in memoryStore)
        self.assertTrue('foo-3' in fileStore and 'foo-4' in fileStore)



name = 'Webware for Python'

version = (3, 0, 0)

status = 'stable'

requiredPyVersion = (3, 6)

synopsis = (
    "Webware for Python is a time-tested"
    " modular, object-oriented web framework.")

webwareConfig = {
    'examplePages': [
        'Welcome',
        'ShowTime',
        'CountVisits',
        'Error',
        'View',
        'Introspect',
        'Colors',
        'ListBox',
        'Forward',
        'SecureCountVisits',
        'FileUpload',
        'RequestInformation',
        'ImageDemo',
        'DominateDemo',
        'YattagDemo',
        'DBUtilsDemo',
        'AjaxSuggest',
        'JSONRPCClient',
    ]
}



import os
import time
import shutil
import tempfile
import unittest

from MiscUtils import PickleCache as pc


class TestPickleCache(unittest.TestCase):

    def setUp(self):
        self._tempDir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self._tempDir, ignore_errors=True)

    @staticmethod
    def remove(filename):
        try:
            os.remove(filename)
        except OSError:
            pass

    def testTwoIterations(self):
        iterations = 2
        for _iteration in range(iterations):
            self.testOneIteration()

    def testOneIteration(self):
        sourcePath = self._sourcePath = os.path.join(self._tempDir, 'foo.dict')
        picklePath = self._picklePath = pc.PickleCache().picklePath(sourcePath)
        self.remove(picklePath)  # make sure we're clean
        data = self._data = dict(x=1)
        self.writeSource()
        try:
            # test 1: no pickle cache yet
            self.assertTrue(pc.readPickleCache(sourcePath) is None)
            self.writePickle()
            # test 2: correctness
            self.assertEqual(pc.readPickleCache(sourcePath), data)
            # test 3: wrong pickle version
            self.assertTrue(
                pc.readPickleCache(sourcePath, pickleProtocol=1) is None)
            self.writePickle()  # restore
            # test 4: wrong data source
            self.assertTrue(
                pc.readPickleCache(sourcePath, source='notTest') is None)
            self.writePickle()  # restore
            # test 5: wrong Python version
            v = list(pc.versionInfo)
            v[-1] += 1  # increment serial number
            v, pc.versionInfo = pc.versionInfo, tuple(v)
            try:
                self.assertTrue(pc.readPickleCache(sourcePath) is None)
                self.writePickle()  # restore
            finally:
                pc.versionInfo = v
            # test 6: source is newer
            self.remove(picklePath)
            self.writePickle()
            # we have to allow for the granularity of getmtime()
            # (see the comment in the docstring of PickleCache.py)
            time.sleep(1.25)
            self.writeSource()
            self.assertTrue(pc.readPickleCache(sourcePath) is None)
            self.writePickle()  # restore
        finally:
            self.remove(sourcePath)
            self.remove(picklePath)

    def writeSource(self):
        with open(self._sourcePath, 'w') as f:
            f.write(str(self._data))

    def writePickle(self):
        self.assertFalse(os.path.exists(self._picklePath))
        pc.writePickleCache(self._data, self._sourcePath, source='test')
        self.assertTrue(os.path.exists(self._picklePath))



"""NamedValueAccess.py

NamedValueAccess provides functions for accessing Python objects by keys and
named attributes. A 'key' is a single identifier such as 'foo'. A 'name' could
be a key, or a qualified key, such as 'foo.bar.boo'. Names are generally more
convenient and powerful, while the key-oriented function is more efficient and
provide the atomic functionality that the name-oriented function is built upon.

CREDIT

Chuck Esterbrook <echuck@mindspring.com>
Tavis Rudd <tavis@calrudd.com>
"""

from MiscUtils import NoDefault


class NamedValueAccessError(LookupError):
    """General named value access error."""


class ValueForKeyError(NamedValueAccessError):
    """No value for key found error."""


def valueForKey(obj, key, default=NoDefault):
    """Get the value of the object named by the given key.

    This method returns the value with the following precedence:
      1. Methods before non-methods
      2. Attributes before keys (__getitem__)
      3. Public things before private things
         (private being denoted by a preceding underscore)

    Suppose key is 'foo', then this method returns one of these:
      * obj.foo()
      * obj._foo()
      * obj.foo
      * obj._foo
      * obj['foo']
      * default  # only if specified

    If all of these fail, a ValueForKeyError is raised.

    NOTES

      * valueForKey() works on dictionaries and dictionary-like objects.
      * See valueForName() which is a more advanced version of this
        function that allows multiple, qualified keys.
    """

    if obj is None:
        raise TypeError('We do not accept None as object')

    if not isinstance(key, str):
        raise TypeError('We only accept strings for keys')

    attr = None
    unknown = False
    if isinstance(obj, dict):
        if default is NoDefault:
            try:
                return obj[key]
            except KeyError:
                raise ValueForKeyError(key) from None
        else:
            return obj.get(key, default)
    else:
        try:
            cls = obj.__class__
        except AttributeError:
            # happens for classes themselves
            cls = method = None
        else:
            method = getattr(cls, key, None)
        if not method:
            underKey = '_' + key
            method = getattr(cls, underKey, None) if cls else None
            if not method:
                attr = getattr(obj, key, NoDefault)
                if attr is NoDefault:
                    attr = getattr(obj, underKey, NoDefault)
                    if attr is NoDefault and cls is not None:
                        getitem = getattr(cls, '__getitem__', None)
                        if getitem:
                            try:
                                getitem(obj, key)
                            except KeyError:
                                unknown = True

    if not unknown:
        if method:
            return method(obj)
        if attr is not NoDefault:
            return attr

    if default is not NoDefault:
        return default

    raise ValueForKeyError(key)


def valueForName(obj, name, default=NoDefault):
    """Get the value of the object that is named.

    The name can use dotted notation to traverse through a network/graph
    of objects. Since this function relies on valueForKey() for each
    individual component of the name, you should be familiar with the
    semantics of that notation.

    Example: valueForName(obj, 'department.manager.salary')
    """
    for name in name.split('.'):
        obj = valueForKey(obj, name, default)
        if obj is default:
            break
    return obj



"""HTTP servlets"""

from time import gmtime, strftime

from Servlet import Servlet


class HTTPServlet(Servlet):
    """A HTTP servlet.

    HTTPServlet implements the respond() method to invoke methods such as
    respondToGet() and respondToPut() depending on the type of HTTP request.
    Example HTTP request methods are GET, POST, HEAD, etc.
    Subclasses implement HTTP method FOO in the Python method respondToFoo.
    Unsupported methods return a "501 Not Implemented" status.

    Note that HTTPServlet inherits awake() and respond() methods from
    Servlet and that subclasses may make use of these.

    See also: Servlet
    """

    # region Init

    def __init__(self):
        Servlet.__init__(self)
        self._methodForRequestType = {}  # a cache; see respond()

    # endregion Init

    # region Transactions

    def respond(self, transaction):
        """Respond to a request.

        Invokes the appropriate respondToSomething() method depending on the
        type of request (e.g., GET, POST, PUT, ...).
        """
        request = transaction.request()
        httpMethodName = request.method()
        # For GET and HEAD, handle the HTTP If-Modified-Since header:
        # If the object's last modified time equals this header, we're done.
        if httpMethodName in ('GET', 'HEAD'):
            lastMod = self.lastModified(transaction)
            if lastMod:
                lastMod = strftime(
                    '%a, %d %b %Y %H:%M:%S GMT', gmtime(lastMod))
                response = transaction.response()
                response.setHeader('Last-Modified', lastMod)
                envGet = request.environ().get
                ims = envGet('HTTP_IF_MODIFIED_SINCE') or envGet(
                    'IF_MODIFIED_SINCE')
                if ims and ims.partition(';')[0] == lastMod:
                    response.delHeader('Content-Type')
                    response.setStatus(304, 'Not Modified')
                    return
        method = self._methodForRequestType.get(httpMethodName)
        if not method:
            methodName = 'respondTo' + httpMethodName.capitalize()
            method = getattr(self, methodName, self.notImplemented)
            self._methodForRequestType[httpMethodName] = method
        method(transaction)

    @staticmethod
    def notImplemented(trans):
        trans.response().setStatus(501, 'Not Implemented')

    @staticmethod
    def lastModified(_trans):
        """Get time of last modification.

        Return this object's Last-Modified time (as a float),
        or None (meaning don't know or not applicable).
        """
        return None

    def respondToHead(self, trans):
        """Respond to a HEAD request.

        A correct but inefficient implementation.
        """
        res = trans.response()
        w = res.write
        res.write = lambda *args: None
        self.respondToGet(trans)  # pylint: disable=no-member
        res.write = w

    # endregion Transactions



from .ExamplePage import ExamplePage


class FileUpload(ExamplePage):
    """This servlet shows how to handle uploaded files.

    The process is fairly self explanatory. You use a form like the one below
    in the writeContent method. When the form is uploaded, the request field
    with the name you gave to the file selector form item will be an instance
    of the FieldStorage class from the standard Python module "cgi". The key
    attributes of this class are shown in the example below. The most important
    things are filename, which gives the name of the file that was uploaded,
    and file, which is an open file handle to the uploaded file. The uploaded
    file is temporarily stored in a temp file created by the standard module.
    You'll need to do something with the data in this file. The temp file will
    be automatically deleted. If you want to save the data in the uploaded file
    read it out and write it to a new file, database, whatever.
    """

    def title(self):
        return "File Upload Example"

    def writeContent(self):
        self.writeln("<h1>Upload Test</h1>")
        try:
            f = self.request().field('filename')
            contents = f.file.read()
        except Exception:
            output = f'''<p>{self.htmlEncode(self.__doc__)}</p>
<form action="FileUpload" method="post" enctype="multipart/form-data">
<input type="file" name="filename">
<input type="submit" value="Upload File">
</form>'''
        else:
            try:
                contentString = contents.decode('utf-8')
            except UnicodeDecodeError:
                try:
                    contentString = contents.decode('latin-1')
                except UnicodeDecodeError:
                    contentString = contents.decode('ascii', 'replace')
            contentString = self.htmlEncode(contentString.strip())
            output = f'''<h4>Here's the file you submitted:</h4>
<table>
<tr><th>name</th><td><strong>{f.filename}</strong></td></tr>
<tr><th>type</th><td>{f.type}</td></tr>
<tr><th>type_options</th><td>{f.type_options}</td></tr>
<tr><th>disposition</th><td>{f.disposition}</td></tr>
<tr><th>disposition_options</th><td>{f.disposition_options}</td></tr>
<tr><th>headers</th><td>{f.headers}</td></tr>
<tr><th>size</th><td>{len(contents)} bytes</td></tr>
<tr><th style="vertical-align:top">contents</th>
<td><pre style="font-size:small;margin:0pt">{contentString}</pre></td></tr>
</table>'''
        self.writeln(output)



"""Configurable.py

Provides configuration file functionality.
"""

import os
import sys

from MiscUtils import AbstractError, NoDefault

from .Funcs import valueForString


class ConfigurationError(Exception):
    """Error in configuration file."""


class Configurable:
    """Abstract superclass for configuration file functionality.

    Subclasses should override:

      * defaultConfig()  to return a dictionary of default settings
                         such as {'Frequency': 5}

      * configFilename() to return the filename by which users can
                         override the configuration such as 'Pinger.config'

    Subclasses typically use the setting() method, for example:

        time.sleep(self.setting('Frequency'))

    They might also use the printConfig() method, for example:

        self.printConfig()     # or
        self.printConfig(file)

    Users of your software can create a file with the same name as
    configFilename() and selectively override settings. The format of
    the file is a Python dictionary.

    Subclasses can also override userConfig() in order to obtain the
    user configuration settings from another source.
    """

    # region Init

    def __init__(self):
        self._config = None

    # endregion Init

    # region Configuration

    def config(self):
        """Return the configuration of the object as a dictionary.

        This is a combination of defaultConfig() and userConfig().
        This method caches the config.
        """
        if self._config is None:
            self._config = {
                **self.defaultConfig(),
                **self.userConfig(),
                **self.commandLineConfig()}
        return self._config

    def setting(self, name, default=NoDefault):
        """Return the value of a particular setting in the configuration."""
        if default is NoDefault:
            try:
                return self.config()[name]
            except KeyError as e:
                keys = ', '.join(sorted(self.config()))
                raise KeyError(
                    f'{name} not found - config keys are: {keys}') from e
        else:
            return self.config().get(name, default)

    def setSetting(self, name, value):
        """Set a particular configuration setting."""
        self.config()[name] = value

    def hasSetting(self, name):
        """Check whether a configuration setting has been changed."""
        return name in self.config()

    def defaultConfig(self):
        """Return a dictionary with all the default values for the settings.

        This implementation returns {}. Subclasses should override.
        """
        return {}

    def configFilename(self):
        """Return the full name of the user config file.

        Users can override the configuration by this config file.
        Subclasses must override to specify a name.
        Returning None is valid, in which case no user config file
        will be loaded.
        """
        raise AbstractError(self.__class__)

    def configName(self):
        """Return the name of the configuration file without the extension.

        This is the portion of the config file name before the '.config'.
        This is used on the command-line.
        """
        return os.path.splitext(os.path.basename(self.configFilename()))[0]

    def configReplacementValues(self):
        """Return a dictionary for substitutions in the config file.

        This must be a dictionary suitable for use with "string % dict"
        that should be used on the text in the config file.
        If an empty dictionary (or None) is returned, then no substitution
        will be attempted.
        """
        return {}

    def userConfig(self):
        """Return the user config overrides.

        These settings can be found in the optional config file.
        Returns {} if there is no such file.

        The config filename is taken from configFilename().
        """
        # pylint: disable=assignment-from-no-return
        filename = self.configFilename()
        if not filename:
            return {}
        try:
            contents = self.readConfig(filename)
        except IOError as e:
            print('WARNING: Config file', filename, 'not loaded:', e.strerror)
            print()
            return {}
        if contents.lstrip().startswith('{'):
            raise ConfigurationError(
                'Configuration via a dict literal is not supported anymore.')
        try:
            from ImportManager import ImportManager
            ImportManager().watchFile(filename)
        except Exception as e:
            print('WARNING: Config file', filename, 'cannot be watched:', e)
            print()
        config = self.configReplacementValues().copy()
        try:
            exec(contents, config)
            keys = [key for key in config if key.startswith('_')]
            for key in keys:
                del config[key]
        except Exception as e:
            raise ConfigurationError(
                f'Invalid configuration file, {filename} ({e}).') from e
        return config

    @staticmethod
    def readConfig(filename):
        """Read the configuration from the file with the given name.

        Raises an UIError if the configuration cannot be read.

        This implementation assumes the file is stored in utf-8 encoding with
        possible BOM at the start, but also tries to read as latin-1 if it
        cannot be decoded as utf-8. Subclasses can override this behavior.
        """
        try:
            with open(filename, encoding='utf-8-sig') as f:
                return f.read()
        except UnicodeDecodeError:
            with open(filename, encoding='latin-1') as f:
                return f.read()

    def printConfig(self, dest=None):
        """Print the configuration to the given destination.

        The default destination is stdout. A fixed with font is assumed
        for aligning the values to start at the same column.
        """
        if dest is None:
            dest = sys.stdout
        config = self.config()
        width = max(map(len, config))
        for key in sorted(config):
            dest.write(f'{key.ljust(width)} = {self.setting(key)!r}\n')
        dest.write('\n')

    def commandLineConfig(self):
        """Return the settings that came from the command-line.

        These settings come via addCommandLineSetting().
        """
        return _settings.get(self.configName(), {})

    # endregion Configuration


# region Command line settings

_settings = {}


def addCommandLineSetting(name, value):
    """Override the configuration with a command-line setting.

    Take a setting, like "Application.Verbose=0", and call
    addCommandLineSetting('Application.Verbose', '0'), and
    it will override any settings in Application.config
    """
    configName, settingName = name.split('.', 1)
    value = valueForString(value)
    if configName not in _settings:
        _settings[configName] = {}
    _settings[configName][settingName] = value


def commandLineSetting(configName, settingName, default=NoDefault):
    """Retrieve a command-line setting.

    You can use this with non-existent classes, like "Context.Root=/Webware",
    and then fetch it back with commandLineSetting('Context', 'Root').
    """
    if default is NoDefault:
        return _settings[configName][settingName]
    return _settings.get(configName, {}).get(settingName, default)

# endregion Command line settings
